from __future__ import annotations
from typing import List

import fusion
from fusion import fsm
from fusion.libs.entity import Entity, get_entity_id
import pamet

from fusion.util import Point2D, Rectangle
from fusion.libs.action import action
from pamet.util import snap_to_grid
from pamet.util.url import Url
from pamet.model.page import Page
from pamet.model.note import Note
from pamet.actions import tab as tab_actions
from pamet.model.arrow import Arrow, ArrowAnchorType
from pamet.model.card_note import CardNote
from pamet.model.image_note import ImageNote
from pamet.model.text_note import TextNote
from pamet.views.arrow.widget import ArrowViewState
from pamet.views.map_page.properties_widget import MapPagePropertiesViewState
from pamet.views.map_page.state import MapPageMode, MapPageViewState
from pamet.views.note.base.state import NoteViewState
from pamet.views.note.qt_helpers import minimal_nonelided_size

log = fusion.get_logger(__name__)


@action('map_page.start_mouse_drag_navigation')
def start_mouse_drag_navigation(map_page_view_state: MapPageViewState,
                                mouse_position: Point2D, first_delta: Point2D):
    map_page_view_state.set_mode(MapPageMode.DRAG_NAVIGATION)
    map_page_view_state.drag_navigation_start_position = mouse_position
    map_page_view_state.viewport_position_on_press = \
        map_page_view_state.viewport_center

    fsm.update_state(map_page_view_state)
    mouse_drag_navigation_move(map_page_view_state, first_delta)


@action('map_page.mouse_drag_navigation_move')
def mouse_drag_navigation_move(map_page_view_state: MapPageViewState,
                               mouse_delta: Point2D):
    unprojected_delta = (mouse_delta /
                         map_page_view_state.height_scale_factor())
    new_viewport_center: Point2D = (
        map_page_view_state.viewport_position_on_press + unprojected_delta)

    map_page_view_state.viewport_center = new_viewport_center
    fsm.update_state(map_page_view_state)


@action('map_page.finish_mouse_drag_navigation')
def finish_mouse_drag_navigation(map_page_view_state: MapPageViewState):
    map_page_view_state.set_mode(MapPageMode.NONE)
    fsm.update_state(map_page_view_state)


@action('map_page.update_note_selections')
def update_child_selections(map_page_view_state: MapPageViewState,
                            selection_updates_by_child: dict):

    if not selection_updates_by_child:
        return

    selection_update_count = 0
    for child, selected in selection_updates_by_child.items():

        if (child in map_page_view_state.selected_children and not selected):
            map_page_view_state.selected_children.remove(child)
            selection_update_count += 1

        elif (child not in map_page_view_state.selected_children and selected):
            map_page_view_state.selected_children.add(child)
            selection_update_count += 1

        else:
            log.warning('Redundant entry in selection_updates_by_note_id')

    # If there's only one arrow selected - show its control points
    single_selected_arrow = None
    if len(map_page_view_state.selected_children) == 1:
        (first_selected_child, ) = map_page_view_state.selected_children
        for arrow_vs in map_page_view_state.arrow_view_states:
            if arrow_vs.id == first_selected_child.id:
                single_selected_arrow = arrow_vs

    if single_selected_arrow:
        map_page_view_state.arrow_with_visible_cps = single_selected_arrow
    else:
        map_page_view_state.arrow_with_visible_cps = None

    if selection_update_count > 0:
        fsm.update_state(map_page_view_state)
        # log.info('Updated %s selections' % selection_update_count)
    else:
        log.info('No selections updated out of %s' %
                 selection_updates_by_child)


@action('map_page.clear_child_selection')
def clear_child_selection(map_page_view_state: str):
    selection_updates = {}
    for selected_child in map_page_view_state.selected_children:
        selection_updates[selected_child] = False

    if not selection_updates:
        return

    update_child_selections(map_page_view_state, selection_updates)


@action('map_page.set_viewport_height')
def set_viewport_height(map_page_view_state: MapPageViewState,
                        new_height: float):
    map_page_view_state.viewport_height = new_height
    fsm.update_state(map_page_view_state)
    # //glutPostRedisplay(); artefact, thank you for siteseeing


@action('map_page.start_drag_select')
def start_drag_select(map_page_view_state: MapPageViewState,
                      position: Point2D):
    map_page_view_state.set_mode(MapPageMode.DRAG_SELECT)
    map_page_view_state.mouse_position_on_drag_select_start = position
    # map_page_view_state.drag_selected_children.update(children_under_mouse)

    fsm.update_state(map_page_view_state)


@action('map_page.update_drag_select')
def update_drag_select(map_page_view_state: MapPageViewState,
                       rect_props: list,
                       drag_selected_children: set = None):
    if drag_selected_children is None:
        drag_selected_children = set()

    map_page_view_state.drag_select_rect_props = rect_props
    map_page_view_state.drag_selected_children.clear()

    for child_state in drag_selected_children:
        if child_state not in map_page_view_state.drag_selected_children:
            map_page_view_state.drag_selected_children.add(child_state)

    fsm.update_state(map_page_view_state)


@action('map_page.stop_drag_select')
def stop_drag_select(map_page_view_state: MapPageViewState, position: Point2D):
    added_selections = (map_page_view_state.drag_selected_children -
                        map_page_view_state.selected_children)
    removed_selections = map_page_view_state.selected_children.intersection(
        map_page_view_state.drag_selected_children)

    selection_updates = {child_state: True for child_state in added_selections}
    selection_updates.update(
        {child_state: False
         for child_state in removed_selections})

    # If there was no drag selection, but only a click

    update_child_selections(map_page_view_state, selection_updates)
    map_page_view_state.clear_mode()

    fsm.update_state(map_page_view_state)


@action('map_page.delete_notes_and_connected_arrows')
def delete_notes_and_connected_arrows(notes: List[Note]):
    if not notes:
        return

    note_ids = []
    page_id = None
    for note in notes:
        note_ids.append(note.own_id)
        if not page_id:
            page_id = note.page_id
        elif page_id != note.page_id:
            raise Exception

    arrows_for_deletion = []
    for arrow in pamet.arrows(pamet.page(note.page_id)):
        if (arrow.tail_note_id and arrow.tail_note_id in note_ids) or\
                (arrow.head_note_id and arrow.head_note_id in note_ids):
            arrows_for_deletion.append(arrow)

    for arrow in arrows_for_deletion:
        pamet.remove_arrow(arrow)

    for note in notes:
        pamet.remove_note(note)


@action('map_page.delete_selected_children')
def delete_selected_children(map_page_view_state: MapPageViewState):
    # Parse the ids into lists for removal, since we need to do it in order -
    # .. ?
    arrows_for_removal = []
    notes_for_removal = []
    for child_state in map_page_view_state.selected_children:
        if isinstance(child_state, NoteViewState):
            note = child_state.get_note()
            notes_for_removal.append(note)
        elif isinstance(child_state, ArrowViewState):
            arrows_for_removal.append(child_state.get_arrow())
        else:
            raise Exception('Unexpected state type')

    # Execute the removal
    for arrow in arrows_for_removal:
        pamet.remove_arrow(arrow)

    delete_notes_and_connected_arrows(notes_for_removal)

    clear_child_selection(map_page_view_state)
    fsm.update_state(map_page_view_state)


@action('map_page.start_notes_resize')
def start_notes_resize(map_page_view_state: MapPageViewState, main_note: Note,
                       mouse_position: Point2D,
                       resize_circle_center_projected: Point2D):
    map_page_view_state.set_mode(MapPageMode.NOTE_RESIZE)
    map_page_view_state.note_resize_delta_from_note_edge = (
        resize_circle_center_projected - mouse_position)
    map_page_view_state.note_resize_click_position = mouse_position
    map_page_view_state.note_resize_main_note = main_note.copy()

    for child_state in map_page_view_state.selected_children:
        if isinstance(child_state, NoteViewState):
            map_page_view_state.note_resize_states.append(child_state)

    # map_page_view_state.note_resize_active = True
    fsm.update_state(map_page_view_state)


@action('map_page.resize_note_views')
def resize_note_views(map_page_view_state: MapPageViewState,
                      new_size: Point2D):
    for ncs in map_page_view_state.note_resize_states:
        ncs.set_size(snap_to_grid(new_size))
        fsm.update_state(ncs)


@action('map_page.finish_notes_resize')
def finish_notes_resize(map_page_view_state: MapPageViewState, new_size: list):
    for note_state in map_page_view_state.note_resize_states:
        note = note_state.get_note()
        note.set_size(snap_to_grid(new_size))
        pamet.update_note(note)
    map_page_view_state.clear_mode()

    fsm.update_state(map_page_view_state)


@action('map_page.start_child_move')
def start_child_move(map_page_view_state: MapPageViewState,
                     mouse_pos: Point2D):
    map_page_view_state.set_mode(MapPageMode.CHILD_MOVE)
    map_page_view_state.mouse_position_on_note_drag_start = mouse_pos
    for child_state in map_page_view_state.selected_children:
        if isinstance(child_state, ArrowViewState):
            map_page_view_state.moved_arrow_states.append(child_state)
        elif isinstance(child_state, NoteViewState):
            map_page_view_state.moved_note_states.append(child_state)
        else:
            raise Exception('Unexpected state type')

    fsm.update_state(map_page_view_state)


@action('map_page.moved_child_view_update')
def moved_child_view_update(map_page_view_state: MapPageViewState,
                            delta: Point2D):
    moved_note_ids = []
    for note_state in map_page_view_state.moved_note_states:
        note = note_state.get_note()
        moved_note_ids.append(note.own_id)
        rect = note.rect()
        rect.set_top_left(snap_to_grid(rect.top_left() + delta))
        note_state.set_rect(rect)
        fsm.update_state(note_state)

    for arrow_state in map_page_view_state.moved_arrow_states:
        arrow: Arrow = arrow_state.get_arrow()
        if arrow.tail_point:
            arrow_state.tail_point = snap_to_grid(arrow.tail_point + delta)
        if arrow.head_point:
            arrow_state.head_point = snap_to_grid(arrow.head_point + delta)

        tail_moved = arrow.tail_point or arrow.tail_note_id in moved_note_ids
        head_moved = arrow.head_point or arrow.head_note_id in moved_note_ids
        if tail_moved and head_moved:
            mid_points = arrow.mid_points
            mid_points = [
                snap_to_grid(mid_point + delta) for mid_point in mid_points
            ]
            arrow_state.replace_midpoints(mid_points)

        fsm.update_state(arrow_state)


@action('map_page.finish_child_move')
def finish_child_move(map_page_view_state: MapPageViewState, delta: Point2D):
    moved_note_ids = []
    for note_state in map_page_view_state.moved_note_states:
        note = note_state.get_note()
        moved_note_ids.append(note.own_id)
        note.x = snap_to_grid(note.x + delta.x())
        note.y = snap_to_grid(note.y + delta.y())
        pamet.update_note(note)

    for arrow_state in map_page_view_state.moved_arrow_states:
        arrow = arrow_state.get_arrow()
        if arrow.tail_point:
            arrow.tail_point = snap_to_grid(arrow.tail_point + delta)
        if arrow.head_point:
            arrow.head_point = snap_to_grid(arrow.head_point + delta)

        tail_moved = arrow.tail_point or arrow.tail_note_id in moved_note_ids
        head_moved = arrow.head_point or arrow.head_note_id in moved_note_ids
        if tail_moved and head_moved:
            mid_points = arrow.mid_points
            mid_points = [
                snap_to_grid(mid_point + delta) for mid_point in mid_points
            ]
            arrow.replace_midpoints(mid_points)
        pamet.update_arrow(arrow)

    map_page_view_state.clear_mode()
    fsm.update_state(map_page_view_state)


@action('map_page.select_all_children')
def select_all_children(map_page_view_state: MapPageViewState):
    for note_vs in map_page_view_state.note_view_states:
        map_page_view_state.selected_children.add(note_vs)

    for arrow_vs in map_page_view_state.arrow_view_states:
        map_page_view_state.selected_children.add(arrow_vs)

    fsm.update_state(map_page_view_state)


@action('map_page.resize_page')
def resize_page(map_page_view_state: MapPageViewState, new_size: Point2D):
    map_page_view_state.geometry.set_size(new_size)
    fsm.update_state(map_page_view_state)


@action('notes.color_selected_children')
def color_selected_children(map_page_view_state: str,
                            color: list = None,
                            background_color: list = None):
    for child_state in map_page_view_state.selected_children:

        if isinstance(child_state, NoteViewState):
            note = child_state.get_note()

            color = color or note.color
            background_color = background_color or note.background_color

            note.color = color
            note.background_color = background_color
            pamet.update_note(note)

        elif isinstance(child_state, ArrowViewState):
            arrow = child_state.get_arrow()
            color = color or arrow.color
            arrow.color = color
            pamet.update_arrow(arrow)
        else:
            raise Exception

    clear_child_selection(map_page_view_state)
    fsm.update_state(map_page_view_state)


@action('map_page.open_page_properties')
def open_page_properties(tab_state: TabViewState, focused_prop: str = None):
    page_view_state = tab_state.page_view_state

    properties_view_state = MapPagePropertiesViewState(
        id=page_view_state.page_id)

    fsm.add_state(properties_view_state)

    if focused_prop:
        properties_view_state.focused_prop = focused_prop
        fsm.update_state(properties_view_state)

    tab_state.page_edit_view_state = properties_view_state

    fsm.update_state(tab_state)


@action('map_page.save_page_properties')
def save_page_properties(page: Page):
    pamet.update_page(page)


@action('map_page.delete_page')
def delete_page(tab_view_state: TabViewState, page: Page):
    pamet.remove_page(page)
    # Try to go to the last page if any
    next_page = None
    if tab_view_state.navigation_history:
        next_page = pamet.page(
            Url(tab_view_state.navigation_history[-1]).get_page_id())

    if not next_page:
        next_page = pamet.default_page() or pamet.sync_repo().find_one(
            type=Page)

    if not next_page:
        next_page = pamet.actions.other.create_default_page()
    tab_actions.go_to_url(tab_view_state, next_page.url())


@action('map_page.handle_child_added', issuer='entity_change_apply_logic')
def handle_child_added(page_view_state: MapPageViewState, child: Entity):
    if isinstance(child, Note):
        if page_view_state.view_state_for_note_own_id(child.own_id):
            return  # A workaround for duplicate calls.
            # These methods should be replaced by a PageUpdate service
            # (instead of per-page subscriptions, etc)
        ViewType = pamet.note_view_type(note_type_name=type(child).__name__,
                                        edit=False)
        StateType = pamet.note_state_type_by_view(ViewType.__name__)
        note_props = child.asdict()
        note_id = note_props.pop('id')
        nv_state = StateType(id=note_id, **note_props)
        fsm.add_state(nv_state)
        page_view_state.note_view_states.add(nv_state)
    elif isinstance(child, Arrow):
        arrow_props = child.asdict()
        arrow_id = arrow_props.pop('id')
        arrow_view_state = ArrowViewState(id=arrow_id, **arrow_props)
        fsm.add_state(arrow_view_state)
        page_view_state.arrow_view_states.add(arrow_view_state)

    fsm.update_state(page_view_state)


@action('map_page.handle_child_removed', issuer='entity_change_apply_logic')
def handle_child_removed(page_view_state: MapPageViewState, child: Entity):
    if isinstance(child, Note):
        nv_states = list(
            filter(lambda x: x.note_gid == child.gid(),
                   page_view_state.note_view_states))
        for nv_state in nv_states:  # Should be len==1
            fsm.remove_state(nv_state)
            page_view_state.note_view_states.remove(nv_state)
    elif isinstance(child, Arrow):
        av_states = list(
            filter(lambda x: x.arrow_gid == child.gid(),
                   page_view_state.arrow_view_states))
        for nv_state in av_states:  # Should be len==1
            fsm.remove_state(nv_state)
            page_view_state.arrow_view_states.remove(nv_state)

    fsm.update_state(page_view_state)


@action('map_page.handle_child_updated', issuer='entity_change_apply_logic')
def handle_child_updated(child_view_state: NoteViewState, child: Note):
    if isinstance(child, Note):
        child_view_state.update_from_note(child)
    elif isinstance(child, Arrow):
        child_view_state.update_from_arrow(child)
    fsm.update_state(child_view_state)


@action('map_page.handle_page_name_updated',
        issuer='entity_change_apply_logic')
def handle_page_name_updated(tab_view_state: TabViewState, page: Page):
    tab_view_state.title = page.name
    fsm.update_state(tab_view_state)


@action('map_page.start_arrow_creation')
def start_arrow_creation(map_page_view_state: MapPageViewState):
    av_state = ArrowViewState().in_page(page_id=map_page_view_state.page_id)
    fsm.add_state(av_state)

    map_page_view_state.set_mode(MapPageMode.CREATE_ARROW)
    map_page_view_state.new_arrow_view_states.append(av_state)
    fsm.update_state(map_page_view_state)


@action('map_page.abort_special_mode')
def abort_special_mode(map_page_view_state: MapPageViewState):
    # Reset the manipulated views
    # For child moving
    for note_state in map_page_view_state.moved_note_states:
        note_state.update_from_note(note_state.get_note())
        fsm.update_state(note_state)

    for arrow_state in map_page_view_state.moved_arrow_states:
        arrow_state.update_from_arrow(arrow_state.get_arrow())
        fsm.update_state(arrow_state)

    # For note resizing
    for note_state in map_page_view_state.note_resize_states:
        note_state.update_from_note(note_state.get_note())
        fsm.update_state(note_state)

    # Restore the arrow view if there was edge dragging
    if map_page_view_state.mode() == MapPageMode.ARROW_EDGE_DRAG:
        map_page_view_state.arrow_with_visible_cps.update_from_arrow(
            map_page_view_state.arrow_with_visible_cps.get_arrow())

    map_page_view_state.clear_mode()
    fsm.update_state(map_page_view_state)


@action('map_page.place_arrow_view_tail')
def place_arrow_view_tail(arrow_view_state: ArrowViewState,
                          fixed_pos: Point2D,
                          anchor_note_id: str = None,
                          anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
    arrow_view_state.set_tail(fixed_pos, anchor_note_id, anchor_type)
    fsm.update_state(arrow_view_state)


@action('map_page.place_arrow_view_head')
def place_arrow_view_head(arrow_view_state: ArrowViewState,
                          fixed_pos: Point2D,
                          anchor_note_id: str = None,
                          anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
    arrow_view_state.set_head(fixed_pos, anchor_note_id, anchor_type)
    fsm.update_state(arrow_view_state)


@action('map_page.arrow_creation_click')
def arrow_creation_click(map_page_view_state: MapPageViewState,
                         fixed_pos: Point2D = None,
                         anchor_note_id: str = None,
                         anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
    should_finish = False

    for arrow_vs in map_page_view_state.new_arrow_view_states:
        if not (arrow_vs.has_tail_anchor() or arrow_vs.tail_point):
            place_arrow_view_tail(arrow_vs, fixed_pos, anchor_note_id,
                                  anchor_type)
            if should_finish:
                raise Exception(
                    'Some of the arrows have tails while others don\'t')
        else:
            place_arrow_view_head(arrow_vs, fixed_pos, anchor_note_id,
                                  anchor_type)
            should_finish = True

    if should_finish:
        finish_arrow_creation(map_page_view_state)


@action('map_page.arrow_creation_move')
def arrow_creation_move(map_page_view_state: MapPageViewState,
                        fixed_pos: Point2D = None,
                        anchor_note_id: str = None,
                        anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
    for av_state in map_page_view_state.new_arrow_view_states:
        place_arrow_view_head(av_state, fixed_pos)


@action('map_page.finish_arrow_creation')
def finish_arrow_creation(map_page_view_state: MapPageViewState):
    for arrow_vs in map_page_view_state.new_arrow_view_states:
        # If both ends of the arrow are at the same point - cancel it
        if arrow_vs.tail_point and arrow_vs.head_point:
            if arrow_vs.tail_point == arrow_vs.head_point:
                continue

        # Similarly if both ends are on the same anchor of the same note -
        # cancel it
        if arrow_vs.has_tail_anchor() and arrow_vs.has_head_anchor():
            if (arrow_vs.tail_note_id == arrow_vs.head_note_id
                    and arrow_vs.tail_anchor == arrow_vs.head_anchor):
                continue

        # Get the properties for the new arrow from the view state
        arrow_dict = arrow_vs.asdict()
        # The new arrow should have a new id in order to avoid
        # conflicts while the mock arrow is still present
        arrow_dict.pop('id')
        arrow = Arrow.in_page(page_id=arrow_vs.page_id)
        # Since the view state has extra attributes
        arrow.replace_silent(**arrow_dict)
        pamet.insert_arrow(arrow)

    map_page_view_state.clear_mode()
    fsm.update_state(map_page_view_state)


@action('map_page.start_arrow_edge_drag')
def start_arrow_edge_drag(map_page_view_state: MapPageViewState,
                          mouse_pos: Point2D, edge_index: float):
    map_page_view_state.dragged_edge_index = edge_index
    map_page_view_state.set_mode(MapPageMode.ARROW_EDGE_DRAG)
    fsm.update_state(map_page_view_state)


@action('map_page.arrow_edge_drag_update')
def arrow_edge_drag_update(
        map_page_view_state: MapPageViewState,
        fixed_pos: Point2D = None,
        anchor_note_id: str = None,
        anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
    arrow_vs: ArrowViewState = map_page_view_state.arrow_with_visible_cps

    if map_page_view_state.dragged_edge_index == 0:
        place_arrow_view_tail(arrow_vs, fixed_pos, anchor_note_id, anchor_type)

    elif map_page_view_state.dragged_edge_index == arrow_vs.edge_indices()[-1]:
        place_arrow_view_head(arrow_vs, fixed_pos, anchor_note_id, anchor_type)

    elif map_page_view_state.dragged_edge_index in arrow_vs.edge_indices(
    ):  # Mid points
        mid_point_idx = int(map_page_view_state.dragged_edge_index - 1)
        mid_points = arrow_vs.mid_points
        mid_points[mid_point_idx] = fixed_pos
        arrow_vs.replace_midpoints(mid_points)

    else:  # .5 indices, so a potential edge, that we must create
        arrow = arrow_vs.get_arrow()
        mid_point_idx = int(map_page_view_state.dragged_edge_index // 1)
        mid_points = arrow.mid_points
        mid_points.insert(mid_point_idx, fixed_pos)
        arrow.replace_midpoints(mid_points)

        arrow_vs.update_from_arrow(arrow)  # Just to be sure
        pamet.update_arrow(arrow)

        map_page_view_state.dragged_edge_index = 1 + mid_point_idx
        fsm.update_state(map_page_view_state)
    fsm.update_state(arrow_vs)


@action('map_page.finish_arrow_edge_drag')
def finish_arrow_edge_drag(
        map_page_view_state: MapPageViewState,
        fixed_pos: Point2D = None,
        anchor_note_id: str = None,
        anchor_type: ArrowAnchorType = ArrowAnchorType.NONE):
    arrow = map_page_view_state.arrow_with_visible_cps.get_arrow()

    if map_page_view_state.dragged_edge_index == 0:
        arrow.set_tail(fixed_pos, anchor_note_id, anchor_type)

    elif map_page_view_state.dragged_edge_index == arrow.edge_indices()[-1]:
        arrow.set_head(fixed_pos, anchor_note_id, anchor_type)

    elif map_page_view_state.dragged_edge_index in arrow.edge_indices(
    ):  # Mid points
        mid_point_idx = int(map_page_view_state.dragged_edge_index - 1)
        mid_points = arrow.mid_points
        mid_points[mid_point_idx] = fixed_pos
        arrow.replace_midpoints(mid_points)

    else:  # .5 indices, so a potential edge, that we must create
        return

    pamet.update_arrow(arrow)
    map_page_view_state.clear_mode()
    fsm.update_state(map_page_view_state)


@action('map_page.delete_control_point')
def delete_arrow_edge(arrow_view_state: ArrowViewState, edge_index: float):
    if edge_index == 0 or edge_index == arrow_view_state.edge_indices()[-1]:
        raise Exception('Cannot remove the tail or head')
    if edge_index not in arrow_view_state.edge_indices():
        raise Exception(f'Cannot remove edge with index {edge_index}')

    arrow = arrow_view_state.get_arrow()
    mid_points = arrow.mid_points
    mid_points.pop(edge_index - 1)
    arrow.replace_midpoints(mid_points)
    pamet.update_arrow(arrow)


@action('map_page.autosize_selected_notes')
def autosize_selected_notes(map_page_view_state: MapPageViewState):
    # views.note.qt_helpers.elide_text should be fixed to not use any Qt stuff
    # in order for this to work on web. The latter only requires abstracting
    # QFontMetrics.boundingRect (that works for single lines) and elideText
    # for single words
    for note_vs in map_page_view_state.selected_children:
        if not isinstance(note_vs, (TextNote, CardNote, ImageNote)):
            continue

        note = note_vs.get_note()
        old_size = note.rect().size()
        new_size = minimal_nonelided_size(note)
        if new_size == old_size:
            continue

        rect = note.rect()
        rect.set_size(snap_to_grid(new_size))
        note.set_rect(rect)
        pamet.update_note(note)


@action('map_page.undo')
def undo(map_page_view_state: MapPageViewState):
    page = pamet.page(map_page_view_state.page_id)
    pamet.undo_service().back_one_step(page.id)


@action('map_page.redo')
def redo(map_page_view_state: MapPageViewState):
    page = pamet.page(map_page_view_state.page_id)
    pamet.undo_service().forward_one_step(page.id)


@action('map_page.copy_selected_children')
def copy_selected_children(map_page_view_state: MapPageViewState,
                           relative_to: Point2D = None):
    copied_notes = []
    positions = []
    for child_vs in map_page_view_state.selected_children:
        # If note view - get note and translate position
        if not isinstance(child_vs, NoteViewState):
            continue

        note: Note = child_vs.get_note()
        positions.append(note.rect().top_left())
        copied_notes.append(note)

    if not copied_notes:
        return

    # If there is no reference point - get the middle of the group
    if not relative_to:
        relative_to = sum(positions) / len(positions)

    # Move the notes relative to the reference point before adding them to the
    # clipboard (since pasting happens relative to the mouse)
    for note in copied_notes:
        rect: Rectangle = note.rect()

        # If there's just one note selected - we don't want a relative pos
        if len(copied_notes) == 1:
            rect.set_top_left(Point2D(0, 0))
        else:
            rect.set_top_left(rect.top_left() - relative_to)
        note.set_rect(rect)

    # If arrow view - get arrow and if not both notes are copied - skip
    # else translate and add to the list
    copied_arrows = []
    for child_vs in map_page_view_state.selected_children:
        # If note view - get note and translate position
        if not isinstance(child_vs, ArrowViewState):
            continue

        arrow: Arrow = child_vs.get_arrow()

        # Check for tail and head notes - if not present - don't copy the arrow
        if arrow.has_head_anchor():
            if arrow.head_note_id not in (note.own_id for note in copied_notes):
                continue
        else:  # If it's a fixed pos - translate it relative to relative_to
            arrow.set_head(fixed_pos=arrow.head_point - relative_to)

        if arrow.has_tail_anchor():
            if arrow.tail_note_id not in (note.own_id for note in copied_notes):
                continue
        else:
            arrow.set_tail(fixed_pos=arrow.tail_point - relative_to)

        mid_points = arrow.mid_points
        mid_points = [mid_point - relative_to for mid_point in mid_points]
        arrow.replace_midpoints(mid_points)

        copied_arrows.append(arrow)
    pamet.clipboard.set_contents(copied_notes + copied_arrows)


@action('map_page.paste')
def paste(map_page_view_state: MapPageViewState, relative_to: Point2D = None):
    page = pamet.page(map_page_view_state.id)
    entities = pamet.clipboard.get_contents()

    updated_ids = {}
    for note in entities:
        if not isinstance(note, Note):
            continue
        if relative_to:
            rect: Rectangle = note.rect()
            top_left = snap_to_grid(relative_to + rect.top_left())
            rect.set_top_left(top_left)
            note.set_rect(rect)

        # Set the proper page id and ensure that there's no conflicting ids
        new_page_id = page.id
        new_own_id = note.own_id
        # If there's a note with the same own_id in the page - create a new one
        if pamet.find(gid=(new_page_id, note.own_id)):
            new_own_id = get_entity_id()
            updated_ids[note.own_id] = new_own_id
        note = note.with_id(page_id=new_page_id, own_id=new_own_id)
        pamet.insert_note(note)

    for arrow in entities:
        if not isinstance(arrow, Arrow):
            continue

        # Set the proper page id and ensure that there's no conflicting ids
        new_page_id = page.id
        new_own_id = arrow.own_id
        if pamet.find(gid=arrow.gid()):
            new_own_id = get_entity_id()
        arrow = arrow.with_id(page_id=new_page_id, own_id=new_own_id)

        # Where the pasted note ids have been changed - correct them
        # in the arrow anchors. Else for fixed anchors - just place them
        # relative to the mouse
        if arrow.has_tail_anchor():
            if arrow.tail_note_id in updated_ids:
                arrow.tail_note_id = updated_ids[arrow.tail_note_id]
        else:
            if relative_to:
                arrow.tail_point = snap_to_grid(relative_to + arrow.tail_point)

        if arrow.has_head_anchor():
            if arrow.head_note_id in updated_ids:
                arrow.head_note_id = updated_ids[arrow.head_note_id]
        else:
            if relative_to:
                arrow.head_point = snap_to_grid(relative_to + arrow.head_point)

        if relative_to:
            mid_points = arrow.mid_points
            mid_points = [
                snap_to_grid(relative_to + mid_point)
                for mid_point in mid_points
            ]
            arrow.replace_midpoints(mid_points)

        pamet.insert_arrow(arrow)


@action('map_page.cut_selected_children')
def cut_selected_children(map_page_view_state: MapPageViewState,
                          relative_to: Point2D = None):
    copy_selected_children(map_page_view_state, relative_to)
    delete_selected_children(map_page_view_state)


@action('map_page.insert_entities_relative_to')
def insert_entities_relative_to(map_page_view_state: MapPageViewState,
                                entities: List[Entity],
                                relative_to: Point2D = None):
    relative_to = relative_to or map_page_view_state.center()
    for note in entities:
        note = note.with_id(page_id=map_page_view_state.page_id)
        rect = note.rect()

        # Move the notes relative to the disired point
        rect.set_top_left(snap_to_grid(rect.top_left() + relative_to))

        # Insert the note
        note.set_rect(rect)
        pamet.insert_note(note)


@action('map_page.paste_special')
def paste_special(map_page_view_state: MapPageViewState,
                  relative_to: Point2D = None):
    entities = pamet.clipboard.convert_external()
    insert_entities_relative_to(map_page_view_state, entities, relative_to)



from fusion.util import Point2D
from fusion.libs.action import action
import pamet

from pamet.util import generate_page_name
from pamet.model.page import Page
from pamet.model.text_note import TextNote


@action('create_default_page')
def create_default_page():
    # TODO: If the repo is empty - put a help link as a first note
    # possibly with a "double-click to open link"

    page = Page(name=generate_page_name())
    pamet.insert_page(page)

    help_link_note = TextNote.in_page(page)
    help_link_note.text = 'Mock help note'

    rect = help_link_note.rect()
    rect.move_center(Point2D(0, 0))
    help_link_note.set_rect(rect)
    pamet.insert_note(help_link_note)

    return page



from __future__ import annotations
from pathlib import Path
import subprocess
from typing import Tuple
from PySide6.QtCore import QTimer, QUrl, Qt
from PySide6.QtGui import QCursor, QDesktopServices
from PySide6.QtWidgets import QApplication, QFileDialog

import fusion
from fusion.util.point2d import Point2D
from fusion.libs.command import command
from pamet import desktop_app, semantic_search_service
import pamet

from pamet.actions import note as note_actions
from pamet.actions import map_page as map_page_actions
from pamet.actions import tab as tab_actions
from pamet.actions import window as window_actions

from pamet.desktop_app.util import current_window, current_tab
from pamet.util import resource_path
from pamet.views.note.base.state import NoteViewState
from pamet.views.note.image.widget import ImageNoteWidget

log = fusion.get_logger(__name__)


def current_tab_and_page_views() -> Tuple[TabWidget, MapPageWidget]:
    tab = current_tab()
    if not tab:
        raise Exception('There\'s no open tab.')

    page_view = tab.page_view()
    if not page_view:
        raise Exception('No page view.')

    return tab, page_view


@command(title='Create new note')
def create_new_note():
    current_window().current_tab().create_new_note_command()


@command(title='Create new page')
def create_new_page():
    current_window().current_tab().create_new_page_command()


@command(title='Open page properties')
def open_page_properties():
    tab, page_view = current_tab_and_page_views()
    map_page_actions.open_page_properties(tab.state())


@command(title='Edit selected note')
def edit_selected_notes():
    tab, page_view = current_tab_and_page_views()

    for note_state in page_view.state().selected_children:
        if not isinstance(note_state, NoteViewState):
            continue
        note_actions.start_editing_note(tab.state(), note_state.get_note())


@command(title='Show all commands')
def open_command_palette():
    window_actions.open_command_view(current_window().state(), prefix='>')


@command(title='Go to file')
def open_command_palette_go_to_file():
    window_actions.open_command_view(current_window().state())


@command(title='Create an arrow')
def start_arrow_creation():
    _, page_view = current_tab_and_page_views()

    map_page_actions.start_arrow_creation(page_view.state())


@command(title='Resize selected notes to fit text')
def autosize_selected_notes():
    _, page_view = current_tab_and_page_views()
    map_page_actions.autosize_selected_notes(page_view.state())
    page_view.autosize_selected_notes()


@command(title='Undo')
def undo():
    _, page_view = current_tab_and_page_views()
    map_page_actions.undo(page_view.state())


@command(title='Redo')
def redo():
    _, page_view = current_tab_and_page_views()
    map_page_actions.redo(page_view.state())


@command(title='Copy')
def copy():
    _, page_view = current_tab_and_page_views()

    q_mouse_pos = page_view.mapFromGlobal(QCursor.pos())
    mouse_pos = Point2D(q_mouse_pos.x(), q_mouse_pos.y())
    map_page_state = page_view.state()
    unproj_mouse_pos = map_page_state.unproject_point(mouse_pos)
    map_page_actions.copy_selected_children(map_page_state, unproj_mouse_pos)


@command(title='Paste')
def paste():
    _, page_view = current_tab_and_page_views()
    q_mouse_pos = page_view.mapFromGlobal(QCursor.pos())
    mouse_pos = Point2D(q_mouse_pos.x(), q_mouse_pos.y())
    map_page_state = page_view.state()
    unproj_mouse_pos = map_page_state.unproject_point(mouse_pos)
    map_page_actions.paste(map_page_state, unproj_mouse_pos)


@command(title='Cut')
def cut():
    _, page_view = current_tab_and_page_views()
    q_mouse_pos = page_view.mapFromGlobal(QCursor.pos())
    mouse_pos = Point2D(q_mouse_pos.x(), q_mouse_pos.y())
    map_page_state = page_view.state()
    unproj_mouse_pos = map_page_state.unproject_point(mouse_pos)
    map_page_actions.cut_selected_children(map_page_state, unproj_mouse_pos)


@command(title='Paste special')
def paste_special():
    _, page_view = current_tab_and_page_views()
    q_mouse_pos = page_view.mapFromGlobal(QCursor.pos())
    mouse_pos = Point2D(q_mouse_pos.x(), q_mouse_pos.y())
    map_page_state = page_view.state()
    unproj_mouse_pos = map_page_state.unproject_point(mouse_pos)
    map_page_actions.paste_special(map_page_state, unproj_mouse_pos)


@command(title='Show search')
def show_global_search():
    tab_view, _ = current_tab_and_page_views()
    tab_actions.open_global_search(tab_view.state())


@command(title='Show semantic search (experimental')
def show_semantic_search():
    tab_view, _ = current_tab_and_page_views()
    if not semantic_search_service():
        raise Exception('Semantic search is not available.'
                        'Itt\'s either broken or you have not enabled it '
                        'in the repo settings.')
    tab_actions.open_semantic_search(tab_view.state())


@command(title='Close tab')
def close_current_tab():
    window = current_window()
    tab = window.current_tab()
    window_actions.close_tab(window.state(), tab.state())


@command(title='Open user settings (JSON)')
def open_user_settings_json():
    settings_path = desktop_app.user_settings_path()
    QDesktopServices.openUrl(QUrl.fromLocalFile(str(settings_path)))


@command(title='Open repo settings (JSON)')
def open_repo_settings_json():
    settings_path = desktop_app.repo_settings_path(pamet.sync_repo().path)
    QDesktopServices.openUrl(QUrl.fromLocalFile(str(settings_path)))


@command(title='Export as web page')
def export_as_web_page():
    tab_view, page_view = current_tab_and_page_views()

    note_elements = []
    for nt_view in page_view.note_views():
        # Skip images
        if isinstance(nt_view, ImageNoteWidget):
            continue

        nv_state = nt_view.state()
        r = nv_state.rect()

        color_vals = [c * 100 for c in nv_state.color]
        color_str = ', '.join([f'{c}%' for c in color_vals])
        color_str = f'rgba({color_str})'

        bg_color_vals = [c * 100 for c in nv_state.background_color]
        bg_color_str = ', '.join([f'{c}%' for c in bg_color_vals])
        bg_color_str = f'rgba({bg_color_str})'

        note_style = (f'top: {r.y()}; left: {r.x()};'
                      f'width: {r.width()};height: {r.height()};')
        note_style += f' color: {color_str};'
        note_style += f' background: {bg_color_str};'

        note_el_props = {
            'id': nv_state.id,
            'class': 'note',
        }
        if nv_state.url.is_external():
            note_el_props['href'] = nv_state.url
            note_style += f'border: 1px solid {color_str};'

        note_el_props['style'] = note_style

        props_str = ' '.join([f'{k}="{v}"' for k, v in note_el_props.items()])
        # id="{nv_state.id}" class="note" style="{note_style}"
        note_el = f'''
        <a {props_str}>
            {nt_view.displayed_text}
        </a>
        '''
        note_elements.append(note_el)

    note_elements_str = "\n".join(note_elements)
    page_script = resource_path('static_page_src/script.js').read_text()
    page_style = resource_path('static_page_src/style.css').read_text()
    page_str = f'''
    <html>
        <head>
            <style>
            {page_style}
            </style>
        </head>
        <body>
            <div id="mapPage">
{note_elements_str}
            </div>

            <script>
                {page_script}
            </script>
        </body>
    </html>
    '''

    # Hackily present the file chooser modal
    # (otherwise the app crashes when the modal is closed)
    file_name = f'{tab_view.state().title}.html'

    def hackily_ask_for_path_and_save_file():
        file_dialog = QFileDialog(parent=page_view)
        path, _ = file_dialog.getSaveFileName(
            None,
            caption='Choose save path',
            dir=file_name,
            filter='*.html',
        )

        if not path:
            return
        path = Path(path)
        path.write_text(page_str)

    QTimer.singleShot(0, hackily_ask_for_path_and_save_file)


@command(title='Grab screen snippet')
def grab_screen_snippet():
    active_window = current_window()
    if not active_window:
        # Minimize all non-active windows
        # (so that the screenshot doesn't include them. It's a problem on KDE)
        for window in QApplication.instance().topLevelWidgets():
            if window != QApplication.activeWindow():
                window.setWindowState(Qt.WindowMinimized)

    desktop_app.get_app().selector_widget.showFullScreen()



from collections import defaultdict
from typing import Dict, List
from fusion.libs.channel import Channel
from fusion.libs.entity.change import Change
import pamet
from pamet.model.page import Page
from pamet.model.page_child import PageChild

MAX_UNDO_HISTORY_SIZE = 1000


class UndoHistory:

    def __init__(self) -> None:
        self._history: List[Change] = []
        self._position_in_history = 0  # 0 is before any changes
        self.applied_changes = []

    def handle_change_set(self, change_set: List[Change]):
        # Filter out the changes that were made by this service
        regular_changes = []
        undo_changes = []
        for change in change_set:
            if change in self.applied_changes:
                undo_changes.append(change)
                self.applied_changes.remove(change)
            else:
                regular_changes.append(change)

        # Undo changes are expected to be in bulk and any leftovers are a
        # sign of a problem
        if undo_changes and (regular_changes or self.applied_changes):
            raise Exception

        if not regular_changes:
            return

        if (self._history and self._position_in_history <
            (len(self._history) - 1)):
            self._history = self._history[:self._position_in_history + 1]
        self._history.append(regular_changes)

        # Keep the hitory size in check
        if len(self._history) > MAX_UNDO_HISTORY_SIZE:
            self._history.pop(0)

        self._position_in_history = len(self._history)

    def has_undo_back(self):
        return self._position_in_history > 0

    def has_undo_forward(self):
        return self._position_in_history < len(self._history)

    def back_one_step(self):
        if not self._history:
            return

        # Index is at the beginning (before changes)
        if not self.has_undo_back():
            return

        changes_to_revert = self._history[self._position_in_history - 1]
        for change in changes_to_revert:
            reversed_change = change.reversed()
            pamet.apply_change(reversed_change)
            self.applied_changes.append(reversed_change)

        self._position_in_history -= 1

    def forward_one_step(self):
        if not self._history:
            return

        # Index is at the end
        if not self.has_undo_forward():
            return

        changes_to_revert = self._history[self._position_in_history]
        for change in changes_to_revert:
            pamet.apply_change(change)
            self.applied_changes.append(change)

        self._position_in_history += 1


class UndoService:

    def __init__(self, change_sets_channel: Channel):
        self._histories_by_page_id: Dict[str, UndoHistory] = defaultdict(
            UndoHistory)

        self.change_channel = change_sets_channel
        change_sets_channel.subscribe(self.handle_change_set)

    def handle_change_set(self, change_set: List[Change]):
        # Validate the change set and determine the page
        changes_by_page_id = defaultdict(list)
        for change in change_set:
            state = change.last_state()
            if isinstance(state, PageChild):
                changes_by_page_id[state.page_id].append(change)
            elif isinstance(state, Page):
                changes_by_page_id[state.id].append(change)
            else:
                raise Exception

        for page_id, changeset_for_page in changes_by_page_id.items():
            self._histories_by_page_id[page_id].handle_change_set(
                changeset_for_page)

    def has_undo_forward(self, page_id):
        if page_id not in self._histories_by_page_id:
            return False
        else:
            return self._histories_by_page_id[page_id].has_undo_forward()

    def has_undo_back(self, page_id):
        if page_id not in self._histories_by_page_id:
            return False
        else:
            return self._histories_by_page_id[page_id].has_undo_back()

    def back_one_step(self, page_id):
        if page_id not in self._histories_by_page_id:
            return
        self._histories_by_page_id[page_id].back_one_step()

    def forward_one_step(self, page_id):
        if page_id not in self._histories_by_page_id:
            return
        self._histories_by_page_id[page_id].forward_one_step()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'widget.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QHBoxLayout, QPushButton, QSizePolicy,
    QWidget)

class Ui_CornerWidget(object):
    def setupUi(self, CornerWidget):
        if not CornerWidget.objectName():
            CornerWidget.setObjectName(u"CornerWidget")
        CornerWidget.resize(366, 46)
        self.horizontalLayout = QHBoxLayout(CornerWidget)
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.navigationBackButton = QPushButton(CornerWidget)
        self.navigationBackButton.setObjectName(u"navigationBackButton")

        self.horizontalLayout.addWidget(self.navigationBackButton)

        self.navigationToggleButton = QPushButton(CornerWidget)
        self.navigationToggleButton.setObjectName(u"navigationToggleButton")

        self.horizontalLayout.addWidget(self.navigationToggleButton)

        self.navigationForwardButton = QPushButton(CornerWidget)
        self.navigationForwardButton.setObjectName(u"navigationForwardButton")

        self.horizontalLayout.addWidget(self.navigationForwardButton)

        self.menuButton = QPushButton(CornerWidget)
        self.menuButton.setObjectName(u"menuButton")

        self.horizontalLayout.addWidget(self.menuButton)


        self.retranslateUi(CornerWidget)

        QMetaObject.connectSlotsByName(CornerWidget)
    # setupUi

    def retranslateUi(self, CornerWidget):
        CornerWidget.setWindowTitle(QCoreApplication.translate("CornerWidget", u"CornerWidget", None))
        self.navigationBackButton.setText(QCoreApplication.translate("CornerWidget", u"<", None))
#if QT_CONFIG(shortcut)
        self.navigationBackButton.setShortcut(QCoreApplication.translate("CornerWidget", u"Alt+Left", None))
#endif // QT_CONFIG(shortcut)
        self.navigationToggleButton.setText(QCoreApplication.translate("CornerWidget", u"toggle", None))
#if QT_CONFIG(shortcut)
        self.navigationToggleButton.setShortcut(QCoreApplication.translate("CornerWidget", u"Backspace", None))
#endif // QT_CONFIG(shortcut)
        self.navigationForwardButton.setText(QCoreApplication.translate("CornerWidget", u">", None))
#if QT_CONFIG(shortcut)
        self.navigationForwardButton.setShortcut(QCoreApplication.translate("CornerWidget", u"Alt+Right", None))
#endif // QT_CONFIG(shortcut)
        self.menuButton.setText(QCoreApplication.translate("CornerWidget", u"Menu", None))
    # retranslateUi




from fusion.util import Color

SELECTION_OVERLAY_COLOR = Color(1, 1, 0, 0.5)
ALIGNMENT_LINE_LENGTH = 120
LONG_PRESS_TIMEOUT = 0.3

MAX_RENDER_TIME = 0.017  # (s) ~60 fps

MOVE_SPEED = 1

DEFAULT_EYE_HEIGHT = 40

MIN_HEIGHT_SCALE = 0.2
MAX_HEIGHT_SCALE = 200

NO_SCALE_LINE_SPACING = 20

ALIGNMENT_GRID_UNIT = 10
AGU = ALIGNMENT_GRID_UNIT

NOTE_MARGIN = AGU / 2
DEFAULT_NOTE_WIDTH = 32 * AGU
DEFAULT_NOTE_HEIGHT = 16 * AGU
MIN_NOTE_WIDTH = 3 * AGU
MIN_NOTE_HEIGHT = 3 * AGU
MAX_NOTE_WIDTH = 192 * AGU
MAX_NOTE_HEIGHT = 192 * AGU

PREFERRED_TEXT_NOTE_ASPECT_RATIO = 5

MAX_AUTOSIZE_WIDTH = 1280  # 16:9
MAX_AUTOSIZE_HEIGHT = 405

DEFAULT_BG_COLOR = (0, 0, 1, 0.1)
DEFAULT_COLOR = (0, 0, 1, 1)

RESIZE_CIRCLE_RADIUS = 2 * AGU

ARROW_SELECTION_RADIUS = 10
ARROW_EDGE_RAIDUS = 25
DEFAULT_ARROW_THICKNESS = 1.5
ARROW_SELECTION_THICKNESS_DELTA = 3.5
CONTROL_POINT_RADIUS = 15
POTENTIAL_EDGE_RADIUS = CONTROL_POINT_RADIUS * 0.7

MAX_NAVIGATION_HISTORY = 10000



from pathlib import Path
from random import randint
import threading
from time import sleep
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
import requests
from uvicorn import Config, Server
from pamet.desktop_app.config import pamet_data_folder_path
from fusion import get_logger

log = get_logger(__name__)

SECRET_REPLY = {'result': 'svoi'}
DEFAULT_PORT = 11352


def port_is_taken(port: int):
    try:
        requests.get(f'http://localhost:{port}/')  #@IgnoreException
    except requests.ConnectionError:
        return False
    return True


class LocalServer:

    def __init__(self,
                 commands: dict = None,
                 config_dir: Path | str = None):
        threading.Thread.__init__(self)
        self.commands = commands or {}
        self.config_dir = config_dir or pamet_data_folder_path

        self.thread = None
        self.app = FastAPI()
        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=['*'],
            allow_credentials=True,
            allow_methods=['*'],
            allow_headers=['*'],
        )
        self.server = None

    def lock_path(self):
        return self.config_dir / '.local_server.lock'

    def get_port_from_lock_file(self):
        lock_file = self.lock_path()
        if lock_file.exists():
            port = lock_file.read_text()
            return int(port)
        else:
            return None

    def write_port_to_lock_file(self, port: int):
        lock_file = self.lock_path()
        lock_file.write_text(str(port))

    def another_instance_is_running(self):
        port = self.get_port_from_lock_file()

        if port is None:
            return False

        # Check with a request
        try:
            reply = requests.get(f'http://localhost:{port}/koi')  #@IgnoreException
            if reply.json() == SECRET_REPLY:
                return True
        except requests.ConnectionError:
            return False
        return True

    def start(self, port: int = None):
        # We're assuming a check has been made if another server is running
        # So if there's a lock - we're going to overwrite it
        self.lock_path().unlink(missing_ok=True)

        port = port or DEFAULT_PORT
        while port_is_taken(port):
            log.warning(f'Port {port} is taken. Trying another one.')
            port = randint(10024, 65535)
        self.write_port_to_lock_file(port)

        # Start the server in a thread
        self.thread = threading.Thread(target=self._run, args=(port,))
        self.thread.start()

    def _run(self, port: int):

        @self.app.post('/commands/{command_name}/')
        def run_command(command_name: str):
            if command_name in self.commands:
                self.commands[command_name]()

        @self.app.get('/koi/')
        def confirm_is_running():
            return SECRET_REPLY

        config = Config(app=self.app, host='127.0.0.1', port=port)
        self.server = Server(config=config)
        self.server.run()

    def stop(self):
        self.server.should_exit = True
        # self.server.should_exit.set()
        self.thread.join()

        # Remove the lock file
        lock_file = self.lock_path()
        lock_file.unlink()

    @staticmethod
    def send_command(port: int, command_name: str):
        requests.post(f'http://localhost:{port}/commands/{command_name}/')


if __name__ == '__main__':
    server = LocalServer()
    server.start()
    sleep(10)
    server.stop()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'props_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.3.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QFrame, QHBoxLayout,
    QLabel, QLineEdit, QPushButton, QSizePolicy,
    QVBoxLayout, QWidget)

class Ui_ScriptNotePropsWidget(object):
    def setupUi(self, ScriptNotePropsWidget):
        if not ScriptNotePropsWidget.objectName():
            ScriptNotePropsWidget.setObjectName(u"ScriptNotePropsWidget")
        ScriptNotePropsWidget.resize(777, 526)
        ScriptNotePropsWidget.setMinimumSize(QSize(640, 0))
        self.verticalLayout = QVBoxLayout(ScriptNotePropsWidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.createScriptWidget = QWidget(ScriptNotePropsWidget)
        self.createScriptWidget.setObjectName(u"createScriptWidget")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.createScriptWidget.sizePolicy().hasHeightForWidth())
        self.createScriptWidget.setSizePolicy(sizePolicy)
        self.verticalLayout_3 = QVBoxLayout(self.createScriptWidget)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.horizontalLayout_7.setContentsMargins(-1, 0, -1, -1)
        self.label_5 = QLabel(self.createScriptWidget)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_7.addWidget(self.label_5)

        self.setDefaultScriptsFolder = QPushButton(self.createScriptWidget)
        self.setDefaultScriptsFolder.setObjectName(u"setDefaultScriptsFolder")

        self.horizontalLayout_7.addWidget(self.setDefaultScriptsFolder)

        self.configureTemplatesButton = QPushButton(self.createScriptWidget)
        self.configureTemplatesButton.setObjectName(u"configureTemplatesButton")

        self.horizontalLayout_7.addWidget(self.configureTemplatesButton)


        self.verticalLayout_3.addLayout(self.horizontalLayout_7)


        self.verticalLayout.addWidget(self.createScriptWidget)

        self.scriptPropertiesWidget = QWidget(ScriptNotePropsWidget)
        self.scriptPropertiesWidget.setObjectName(u"scriptPropertiesWidget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.scriptPropertiesWidget.sizePolicy().hasHeightForWidth())
        self.scriptPropertiesWidget.setSizePolicy(sizePolicy1)
        self.verticalLayout_2 = QVBoxLayout(self.scriptPropertiesWidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(-1, 0, -1, -1)
        self.runInTerminalCheckbox = QCheckBox(self.scriptPropertiesWidget)
        self.runInTerminalCheckbox.setObjectName(u"runInTerminalCheckbox")
        self.runInTerminalCheckbox.setChecked(True)

        self.horizontalLayout.addWidget(self.runInTerminalCheckbox)

        self.runPushButton = QPushButton(self.scriptPropertiesWidget)
        self.runPushButton.setObjectName(u"runPushButton")

        self.horizontalLayout.addWidget(self.runPushButton)


        self.verticalLayout_2.addLayout(self.horizontalLayout)

        self.line = QFrame(self.scriptPropertiesWidget)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.verticalLayout_2.addWidget(self.line)

        self.openInEditorPushButton = QPushButton(self.scriptPropertiesWidget)
        self.openInEditorPushButton.setObjectName(u"openInEditorPushButton")

        self.verticalLayout_2.addWidget(self.openInEditorPushButton)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_2 = QLabel(self.scriptPropertiesWidget)
        self.label_2.setObjectName(u"label_2")

        self.horizontalLayout_2.addWidget(self.label_2)

        self.scriptPathLineEdit = QLineEdit(self.scriptPropertiesWidget)
        self.scriptPathLineEdit.setObjectName(u"scriptPathLineEdit")

        self.horizontalLayout_2.addWidget(self.scriptPathLineEdit)

        self.pathErrorLabel = QLabel(self.scriptPropertiesWidget)
        self.pathErrorLabel.setObjectName(u"pathErrorLabel")

        self.horizontalLayout_2.addWidget(self.pathErrorLabel)

        self.chooseFilePushButton = QPushButton(self.scriptPropertiesWidget)
        self.chooseFilePushButton.setObjectName(u"chooseFilePushButton")

        self.horizontalLayout_2.addWidget(self.chooseFilePushButton)

        self.deletePushButton = QPushButton(self.scriptPropertiesWidget)
        self.deletePushButton.setObjectName(u"deletePushButton")

        self.horizontalLayout_2.addWidget(self.deletePushButton)


        self.verticalLayout_2.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label = QLabel(self.scriptPropertiesWidget)
        self.label.setObjectName(u"label")

        self.horizontalLayout_3.addWidget(self.label)

        self.scriptNameLineEdit = QLineEdit(self.scriptPropertiesWidget)
        self.scriptNameLineEdit.setObjectName(u"scriptNameLineEdit")

        self.horizontalLayout_3.addWidget(self.scriptNameLineEdit)

        self.renamePushButton = QPushButton(self.scriptPropertiesWidget)
        self.renamePushButton.setObjectName(u"renamePushButton")

        self.horizontalLayout_3.addWidget(self.renamePushButton)

        self.openContainingFolderPushButton = QPushButton(self.scriptPropertiesWidget)
        self.openContainingFolderPushButton.setObjectName(u"openContainingFolderPushButton")

        self.horizontalLayout_3.addWidget(self.openContainingFolderPushButton)


        self.verticalLayout_2.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.label_3 = QLabel(self.scriptPropertiesWidget)
        self.label_3.setObjectName(u"label_3")

        self.horizontalLayout_4.addWidget(self.label_3)

        self.argumentsLineEdit = QLineEdit(self.scriptPropertiesWidget)
        self.argumentsLineEdit.setObjectName(u"argumentsLineEdit")

        self.horizontalLayout_4.addWidget(self.argumentsLineEdit)


        self.verticalLayout_2.addLayout(self.horizontalLayout_4)


        self.verticalLayout.addWidget(self.scriptPropertiesWidget)

        self.noteTextVerticalLayout = QVBoxLayout()
        self.noteTextVerticalLayout.setObjectName(u"noteTextVerticalLayout")
        self.noteTextVerticalLayout.setContentsMargins(-1, 0, -1, -1)
        self.label_4 = QLabel(ScriptNotePropsWidget)
        self.label_4.setObjectName(u"label_4")

        self.noteTextVerticalLayout.addWidget(self.label_4)


        self.verticalLayout.addLayout(self.noteTextVerticalLayout)


        self.retranslateUi(ScriptNotePropsWidget)

        QMetaObject.connectSlotsByName(ScriptNotePropsWidget)
    # setupUi

    def retranslateUi(self, ScriptNotePropsWidget):
        ScriptNotePropsWidget.setWindowTitle(QCoreApplication.translate("ScriptNotePropsWidget", u"ScriptNotePropsWidget", None))
        self.label_5.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Create new:", None))
        self.setDefaultScriptsFolder.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Set default folder", None))
        self.configureTemplatesButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Configure templates", None))
        self.runInTerminalCheckbox.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Run in terminal", None))
        self.runPushButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Run", None))
        self.openInEditorPushButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Open in editor", None))
        self.label_2.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Path/Command", None))
        self.pathErrorLabel.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Path error label", None))
        self.chooseFilePushButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Choose file", None))
        self.deletePushButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Delete", None))
        self.label.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Name", None))
        self.scriptNameLineEdit.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"banana-potato-generated-name.py", None))
        self.renamePushButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Rename", None))
        self.openContainingFolderPushButton.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Open containing folder", None))
        self.label_3.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Arguments", None))
        self.label_4.setText(QCoreApplication.translate("ScriptNotePropsWidget", u"Note text:", None))
    # retranslateUi




from __future__ import annotations
import json
from PySide6.QtCore import Qt, QPoint, QSize
from PySide6.QtGui import QShortcut, QKeySequence
from PySide6.QtWidgets import QMessageBox, QPushButton, QWidget

from fusion.libs.entity import load_from_dict
from fusion.view import View
from pamet.actions import note as note_actions
from pamet.note_view_lib import note_types_with_assiciated_views
from pamet.views.context_menu.widget import ContextMenuWidget
from pamet.views.map_page.widget import MapPageWidget
from pamet.views.note.base_edit.view_state import NoteEditViewState
from pamet.desktop_app import icons

from .ui_widget import Ui_BaseNoteEditWidget


class BaseNoteEditWidget(QWidget, View):

    def __init__(self, parent: TabWidget, initial_state: NoteEditViewState):
        QWidget.__init__(self, parent=parent)
        View.__init__(self, initial_state=initial_state)
        self.parent_tab: TabWidget = parent

        if initial_state.new_note_dict:
            self.edited_note = load_from_dict(
                initial_state.new_note_dict)
        else:
            self.edited_note = initial_state.get_note()

            if not self.edited_note:
                raise Exception

        self.ui = Ui_BaseNoteEditWidget()
        self.ui.setupUi(self)

        trash_button = QPushButton(icons.trash, '', self)
        self.ui.toolbarLayout.addWidget(trash_button)

        self.switch_type_button = QPushButton(icons.more, '', self)
        self.ui.toolbarLayout.addWidget(self.switch_type_button)

        QShortcut(QKeySequence(Qt.Key_Escape), self, self._handle_esc_shortcut)
        self.ui.saveButton.clicked.connect(self._handle_ok_click)
        self.ui.devButton.clicked.connect(self._handle_dev_button_click)
        self.ui.cancelButton.clicked.connect(
            lambda: note_actions.abort_editing_note(self.parent_tab.state()))
        trash_button.clicked.connect(
            lambda: note_actions.abort_editing_and_delete_note(self.parent_tab.
                                                               state()))
        self.switch_type_button.clicked.connect(
            self._show_the_type_switch_menu)

        page_view: MapPageWidget = parent.page_view()
        page_view_state = page_view.state()

        # Center the widget over the note being edited with a size possibly
        # the same as the notes unscaled size
        size = initial_state.rect().size()
        center = page_view_state.project_point(initial_state.rect().center())
        center = page_view.mapToGlobal(QPoint(*center.as_tuple()))

        self_rect = self.geometry()
        delta_y = self.ui.bottomLayout.geometry().size().height()
        delta_y += self.ui.topLayout.geometry().size().height()
        size = QSize(*size.as_tuple())
        size.setHeight(size.height() + delta_y)
        self_rect.setSize(size)
        self_rect.moveCenter(center)
        self.setGeometry(self_rect)

    def _handle_esc_shortcut(self):
        note_actions.abort_editing_note(self.parent_tab.state())

    def _handle_ok_click(self):
        if self.state().create_mode:
            note_actions.finish_creating_note(self.parent_tab.state(),
                                              self.edited_note)
        else:
            note_actions.finish_editing_note(self.parent_tab.state(),
                                             self.edited_note)

    def _handle_dev_button_click(self):
        note = self.state().get_note()
        if note:
            note_dict_str = json.dumps(note.asdict(),
                                       indent=4,
                                       ensure_ascii=False)
        else:
            note_dict_str = 'Creating new note, no info'
        QMessageBox.information(
            self,
            'Note info',
            note_dict_str,
            # buttons=QMessageBox.Ok,
        )

    def _show_the_type_switch_menu(self):
        switch_type_menu_entries = {}
        for NoteType in note_types_with_assiciated_views():

            def command(NoteType=NoteType):
                # Possibly ask for a scripts permission here
                note_actions.switch_note_type(
                    self.parent_tab.state(),
                    NoteType.create_silent(**self.edited_note.asdict()))

            switch_type_menu_entries[NoteType.__name__] = command

        menu_dict = {'Switch type:': None, **switch_type_menu_entries}
        menu = ContextMenuWidget(self, menu_dict)

        # Center the menu under the button and show it
        position = self.switch_type_button.rect().center()
        position.setX(position.x() - menu.width() / 2)
        menu.popup(position)



import pamet
from fusion.visual_inspection.action_camera import exec_action
from fusion import fsm

from pamet.model.text_note import TextNote
from fusion.util.point2d import Point2D


# Resize note
# Move note
# Edit note (with multiline text)
# Delete note
# Create note
def test_note_crud(window_fixture, request):
    run_headless = request.config.getoption('--headless')

    with exec_action(delay_before_next=0.001154184341430664,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('4'): True})

    with exec_action(delay_before_next=0.00896000862121582,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.007747173309326172,
                     apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('4'): True})

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.start_notes_resize(
            map_page_view_state=fsm.view_state('3'),
            main_note=TextNote(id=("00000001", "00000002"),
                               geometry=[-160.0, -80.0, 320, 160],
                               style={},
                               content={"text": "Mock help note"},
                               metadata={},
                               created="2022-09-09T10:03:12+03:00",
                               modified="2022-09-09T10:03:12+03:00",
                               tags=[]),
            mouse_position=Point2D(1002, 523),
            resize_circle_center_projected=Point2D(999.0, 521.5))

    with exec_action(delay_before_next=0.5956573486328125,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_note_views(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(188.0, 95.0))

    with exec_action(delay_before_next=0.5051360130310059,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_note_views(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(186.0, 95.0))

    with exec_action(delay_before_next=0.5053880214691162,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_note_views(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(176.0, 93.0))

    with exec_action(delay_before_next=0.003194570541381836,
                     apply_delay=not run_headless):
        pamet.actions.map_page.resize_note_views(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(174.0, 93.0))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.finish_notes_resize(
            map_page_view_state=fsm.view_state('3'),
            new_size=Point2D(174.0, 93.0))

    with exec_action(delay_before_next=0.003925800323486328,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('4'): True})

    with exec_action(delay_before_next=0.5153989791870117,
                     apply_delay=not run_headless):
        pamet.actions.map_page.start_child_move(
            map_page_view_state=fsm.view_state('3'),
            mouse_pos=Point2D(884, 455))

    with exec_action(delay_before_next=0.5095925331115723,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(-8.0, -74.0))

    with exec_action(delay_before_next=0.5160834789276123,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(-24.0, -178.0))

    with exec_action(delay_before_next=0.5158922672271729,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(-20.0, -210.0))

    with exec_action(delay_before_next=0.007718324661254883,
                     apply_delay=not run_headless):
        pamet.actions.map_page.moved_child_view_update(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(-20.0, -212.0))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.finish_child_move(
            map_page_view_state=fsm.view_state('3'),
            delta=Point2D(-20.0, -212.0))

    with exec_action(delay_before_next=0.0045282840728759766,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.update_child_selections(
            map_page_view_state=fsm.view_state('3'),
            selection_updates_by_child={fsm.view_state('4'): True})

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.start_editing_note(
            tab_view_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000002"),
                          geometry=[-180, -290, 170, 90],
                          style={},
                          content={"text": "Mock help note"},
                          metadata={},
                          created="2022-09-09T10:03:12+03:00",
                          modified="2022-09-09T10:03:12+03:00",
                          tags=[]))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.finish_editing_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000002"),
                          geometry=[-180, -290, 170, 90],
                          style={},
                          content={"text": "Mock help note\nEdited multiline"},
                          metadata={},
                          created="2022-09-09T10:03:12+03:00",
                          modified="2022-09-09T10:03:36+03:00",
                          tags=[]))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.map_page.delete_selected_children(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=0.5052444934844971,
                     apply_delay=not run_headless):
        pamet.actions.map_page.clear_child_selection(
            map_page_view_state=fsm.view_state('3'))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.create_new_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000005"),
                          geometry=[-128.0, -207.0, 320, 160],
                          style={},
                          content={"text": ""},
                          metadata={},
                          created="2022-09-09T10:03:44+03:00",
                          modified="2022-09-09T10:03:44+03:00",
                          tags=[]))

    with exec_action(delay_before_next=1, apply_delay=not run_headless):
        pamet.actions.note.finish_creating_note(
            tab_state=fsm.view_state('2'),
            note=TextNote(id=("00000001", "00000005"),
                          geometry=[-128.0, -207.0, 100, 30],
                          style={},
                          content={"text": "New note"},
                          metadata={},
                          created="2022-09-09T10:03:47+03:00",
                          modified="2022-09-09T10:03:47+03:00",
                          tags=[]))

    assert True



from __future__ import annotations
from copy import copy

from dataclasses import field
from pathlib import Path
from fusion.libs.entity import entity_type, Entity
from fusion.logging import get_logger

log = get_logger(__name__)


@entity_type
class PametSettings(Entity):
    _dict_on_load: dict = field(default_factory=dict, init=False, repr=False)

    @classmethod
    def load(cls: PametSettings, config_dict: dict):
        dict_on_load = copy(config_dict)
        id = config_dict.pop('id', None)
        config = cls(id=id) if id else cls()
        config._dict_on_load = dict_on_load
        leftovers = config.replace_silent(**config_dict)
        if leftovers:
            log.error(f'Unrecognized keys found in the config: {leftovers}')

        return config

    def changes_present(self) -> bool:
        """Returns true if the config has been changed since it has been
        loaded.

        Can return True directly after load since there may have been some
        incompatible manual edit or missing default. The class applies those
        at load time.
        """
        return self._dict_on_load != self.asdict()

    def asdict(self) -> dict:
        self_dict = super().asdict()
        for key, val in self_dict.items():
            if isinstance(val, Path):
                self_dict[key] = str(val)

        return self_dict



from copy import copy
from PySide6.QtWidgets import QApplication
from fusion.libs.entity import Entity
from pamet.desktop_app.mime_data_utils import entities_from_mime_data
from pamet.model.image_note import ImageNote
from pamet.model.text_note import TextNote
from fusion import get_logger

log = get_logger(__name__)


class ClipboardService:

    def __init__(self):
        self._internal: list[Entity] = []

    def set_contents(self, entities: list[Entity]):
        if not entities:
            return

        self._internal = copy(entities)

        # The notes will be converted to markdown one fine day
        ext_clipboard_items = []
        for entity in entities:
            if isinstance(entity, TextNote):
                ext_clipboard_items.append(entity.text)
            elif isinstance(entity, ImageNote):
                ext_clipboard_items.append(str(entity.image_url))

        ext_content = '\n\n'.join(ext_clipboard_items)
        clipboard = QApplication.clipboard()
        clipboard.setText(ext_content)

    def get_contents(self) -> list[Entity]:
        return [entity.copy() for entity in self._internal]

    def convert_external(self) -> list[Entity]:
        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()
        return entities_from_mime_data(mime_data)



from PySide6.QtCore import Qt
from PySide6.QtGui import QShowEvent
from PySide6.QtWidgets import QListWidgetItem, QWidget

from fusion.libs.state import ViewState, view_state_type
import pamet
from pamet.util.url import Url
from pamet.services.search.base import SearchResult
from pamet.actions import tab as tab_actions

from .ui_widget import Ui_SearchBarWidget

MAX_RESULTS_COUNT = 1000


@view_state_type
class SearchBarWidgetState(ViewState):
    pass


class SearchItem(QListWidgetItem):

    def __init__(self, search_result: SearchResult):
        # Limit the result text legnth
        result_text = search_result.content_string
        if len(search_result.content_string) > 100:
            result_text = search_result.content_string[:100] + '...'
            
        QListWidgetItem.__init__(self, result_text)
        self.setData(Qt.UserRole, search_result)
        # self.setBackground(Qt.gray)
        # self.setText(search_result.content_string)


class SearchBarWidget(QWidget):

    def __init__(self, parent, initial_state):
        QWidget.__init__(self, parent)

        self.parent_tab = parent

        self.ui = Ui_SearchBarWidget()
        self.ui.setupUi(self)

        self.ui.searchLineEdit.textChanged.connect(self.update_search_results)
        palette = self.palette()
        self.setStyleSheet(f"""QListWidget::item {{
                border-bottom: 1px solid {palette.mid().color().name()};
            }}""")

        self.ui.resultsListWidget.itemClicked.connect(
            self.handle_item_activated)

    @property
    def search_service(self):
        return pamet.search_service()

    def showEvent(self, event: QShowEvent):
        self.ui.searchLineEdit.setFocus()
        self.ui.searchLineEdit.selectAll()
        return super().showEvent(event)

    def update_search_results(self, search_text):
        self.ui.resultsListWidget.clear()

        if not search_text:
            return

        count = 0
        for result in self.search_service.text_search(search_text):
            self.ui.resultsListWidget.addItem(SearchItem(result))
            count += 1

            if count > MAX_RESULTS_COUNT:
                break

    def handle_item_activated(self, item: SearchItem):
        search_result: SearchResult = item.data(Qt.UserRole)
        note = search_result.get_note()
        page = pamet.page(note.page_id)

        url = Url(page.url()).with_anchor(note=note)
        url = str(url)
        tab_actions.go_to_url(self.parent_tab.state(), url)



from dataclasses import dataclass
from typing import List, Tuple

from fusion.libs.entity.change import Change
from fusion.libs.channel import Channel
import pamet
from pamet.model.note import Note


def content_str_for_note(note):
    content = note.content
    # Disregard empty keys
    content = {key: value for key, value in content.items() if value}
    if len(content) == 1 and 'text' in content:
        # If there's only text: add just the text
        content_str = content['text']
    else:
        # If there's other keys (e.g. 'url', etc) - format them
        content_rows = []
        for key, item in content.items():
            content_rows.append(f'{key}: {item}')
        content_str = '\n'.join(content_rows)
    return content_str


def concat_content_for_note(note: Note) -> str:
    return ' '.join([str(val) for val in note.content.values()])


@dataclass
class SearchResult:
    note_gid: Tuple
    content_string: str
    score: float

    def get_note(self):
        return pamet.find_one(gid=self.note_gid)


class IndexEntry:
    def __init__(self, note: Note):
        self.note_gid = note.gid()
        self.content_string = content_str_for_note(note)
        self.content_lowered = concat_content_for_note(note).lower()


class BaseSearchService:

    def __init__(self, change_set_channel: Channel):
        self.index = {}  # By note gid
        change_set_channel.subscribe(self.handle_change_set)
        # self.ready = False

    def upsert_to_index(self, entry: IndexEntry):
        self.index[entry.note_gid] = entry

    def remove_from_index(self, note_gid: Tuple):
        self.index.pop(note_gid)

    def load_all_content(self):
        for page in pamet.pages():
            for note in pamet.notes(page):
                self.upsert_to_index(IndexEntry(note))
        # self.ready = True

    def handle_change_set(self, change_set: List[Change]):
        """Keeps the cached texts up to date using the received changes."""
        for change in change_set:
            state = change.last_state()
            if not isinstance(state, Note):
                continue

            if change.is_delete():
                if state.gid() in self.index:
                    self.remove_from_index(state.gid())
            else:  # change is_create or is_update
                self.upsert_to_index(IndexEntry(state))

    def text_search(self, text: str) -> SearchResult:
        """Searches the cached texts and returns search results"""
        raise NotImplementedError



from __future__ import annotations
from fusion.libs.entity import Entity, get_entity_class_by_name

_view_types_by_note_type_name = {}
_edit_view_types_by_note_type_name = {}
_all_by_state_type_name = {}
_states_by_view_type_name = {}


def register_note_view_type(state_type, note_type, edit):
    state_type_name = state_type.__name__

    if not isinstance(note_type, (list, tuple)):
        note_type_names = [note_type.__name__]
    else:
        note_type_names = [nt.__name__ for nt in note_type]

    def view_registration_decorator(cls):
        if edit:
            for note_type_name in note_type_names:
                if note_type_name in _edit_view_types_by_note_type_name:
                    raise Exception('Class already registered (edit)')
                _edit_view_types_by_note_type_name[note_type_name] = cls
        else:
            for note_type_name in note_type_names:
                if note_type_name in _view_types_by_note_type_name:
                    raise Exception('Class already registered')
                _view_types_by_note_type_name[note_type_name] = cls

        _all_by_state_type_name[state_type_name] = cls
        _states_by_view_type_name[cls.__name__] = state_type
        return cls

    return view_registration_decorator


def note_view_type(note_type_name, edit=False) -> NoteViewState:
    if edit:
        return _edit_view_types_by_note_type_name[note_type_name]
    else:
        return _view_types_by_note_type_name[note_type_name]


def note_view_type_by_state(state_type_name):
    return _all_by_state_type_name[state_type_name]


def note_state_type_by_view(view_type_name):
    return _states_by_view_type_name[view_type_name]


def note_view_state_type_for_note(note, edit=False):
    NoteViewType = note_view_type(type(note).__name__, edit)
    return note_state_type_by_view(NoteViewType.__name__)


def note_types_with_assiciated_views() -> Entity:
    note_types = []
    for note_type_name in _view_types_by_note_type_name:
        note_type = get_entity_class_by_name(note_type_name)
        note_types.append(note_type)
    return note_types



from fusion.libs.action import action
from fusion.libs.entity.change import Change
from fusion.util.point2d import Point2D
from pamet import channels
import pamet
from pamet.constants import ALIGNMENT_GRID_UNIT
from pamet.model.note import Note
from pamet.model.op_list_note import OtherPageListNote
from pamet.model.page import Page
from pamet.model.text_note import TextNote
from pamet.util import snap_to_grid
from pamet.views.note.qt_helpers import minimal_nonelided_size

DUPLICATE_LIST_NOTE_WARNING = 'Only one OtherPageList note allowed per page.'

# Test cases:
# Add/remove/update link note in a page with an OPList
# Add page


class OtherPagesListUpdateService:

    def __init__(self) -> None:
        self.OPL_notes_by_page_id = {}

    def start(self):
        channels.entity_change_sets_per_TLA.subscribe(
            self.handle_entity_changes)

        # Create the oplist cache
        for note in pamet.find(type=OtherPageListNote):
            if note.page_id in self.OPL_notes_by_page_id:
                self.swap_duplicate_OPLNote_for_warning(note)
                continue
            self.OPL_notes_by_page_id[note.page_id] = note

        self.update_all_oplists()

    @action('swap_duplicate_oplist_note_for_warning')
    def swap_duplicate_OPLNote_for_warning(self, note: OtherPageListNote):
        note = TextNote().with_id(note.id)
        note.text = DUPLICATE_LIST_NOTE_WARNING
        pamet.update_note(note)

    @action('update_all_oplists')
    def update_all_oplists(self):
        for page_id in self.OPL_notes_by_page_id:
            self.update_oplist(page_id)

    @action('update_oplist')
    def update_oplist(self, page_id: str):
        if page_id not in self.OPL_notes_by_page_id:
            return

        oplist_note = self.OPL_notes_by_page_id[page_id]

        # Get all the other notes in the page. Make a set with the ones linking
        # to internal pages and fill in the OPList to hold links to all pages
        link_notes_by_page_id = {}
        all_page_ids = [page.id for page in pamet.pages()]
        notes = pamet.notes(page_id)
        notes_probably_in_list = []
        list_left_border = oplist_note.rect().x() + ALIGNMENT_GRID_UNIT
        for note in notes:
            if note.rect().x() == list_left_border:
                notes_probably_in_list.append(note)

            if not note.url.is_internal():
                continue
            link_notes_by_page_id[note.url.get_page_id()] = note

        missing_links = set(all_page_ids) - set(link_notes_by_page_id)
        spawn_pos = oplist_note.rect().bottom_left() + Point2D(
            ALIGNMENT_GRID_UNIT, ALIGNMENT_GRID_UNIT)

        # Find the bottom of the list in order to add the note there
        probe_spawn_pos = spawn_pos
        while probe_spawn_pos and notes_probably_in_list:
            found_next = False
            for note in notes_probably_in_list:
                rect = note.rect()
                if rect.contains(probe_spawn_pos):
                    probe_spawn_pos = (rect.bottom_left() +
                                       Point2D(0, ALIGNMENT_GRID_UNIT))
                    found_next = True
                    break

            if not found_next:
                spawn_pos = probe_spawn_pos
                probe_spawn_pos = None
                break

        for missing_page_id in missing_links:
            new_note = TextNote.in_page(page_id=page_id)
            new_note.url = pamet.page(missing_page_id).url()

            # Move to the initial spawn position
            rect = new_note.rect()
            rect.set_top_left(spawn_pos)

            # Autosize
            new_size = minimal_nonelided_size(new_note)
            rect.set_size(snap_to_grid(new_size))

            # Move to the spawn pos
            rect.set_top_left(spawn_pos)

            new_note.set_rect(rect)
            pamet.insert_note(new_note)

            # Move the spawn pos
            spawn_pos.set_y(rect.bottom() + ALIGNMENT_GRID_UNIT)

    def handle_entity_changes(self, change_set: list[Change]):
        pages_for_update = set()

        for change in change_set:
            entity = change.last_state()

            if isinstance(entity, Note):
                # If its an OPList note - keep the cache up to date
                if isinstance(entity, OtherPageListNote):
                    if change.is_create():
                        # If there's already an OPList note in that page -
                        # disable it
                        if entity.page_id in self.OPL_notes_by_page_id:
                            self.swap_duplicate_OPLNote_for_warning(entity)
                            continue

                        self.OPL_notes_by_page_id[entity.page_id] = entity
                        pages_for_update.add(entity.page_id)
                    elif change.is_delete():
                        self.OPL_notes_by_page_id.pop(entity.page_id)

                # If the note change is in one of the pages with an OPList
                # if it's a change in the internal link - update the OPList
                elif entity.page_id in self.OPL_notes_by_page_id:
                    if ((entity.url.is_internal()) and
                        (change.is_create() or change.is_delete())
                            or (change.is_update() and change.updated.url)):
                        pages_for_update.add(entity.page_id)

                for page_id in pages_for_update:
                    self.update_oplist(page_id)

            elif isinstance(entity, Page) and change.is_create():
                self.update_all_oplists()



import fusion
from fusion.util import Point2D, Rectangle


class Viewport:
    # def __init__(self):
    #     self =

    def __repr__(self):
        info = (self.center(), self.eye_height)
        return '<Viewport center=%s eyeHeight=%s>' % info

    def center(self) -> Point2D:
        return self.viewport_center

    @property
    def eye_height(self) -> float:
        return self.viewport_height

    def height_scale_factor(self):
        return fusion.line_spacing_in_pixels / self.eye_height

    def project_rect(self, rect: Rectangle) -> Rectangle:
        top_left = self.project_point(rect.top_left())
        bottom_right = self.project_point(rect.bottom_right())
        return Rectangle.from_points(top_left, bottom_right)

    def project_point(self, point: Point2D) -> Point2D:
        return Point2D(self.project_x(point.x()), self.project_y(point.y()))

    def project_x(self, x_on_page: float) -> float:
        x_on_page -= self.center().x()
        x_on_page *= self.height_scale_factor()
        return x_on_page + self.geometry.width() / 2

    def project_y(self, y_on_page: float) -> float:
        y_on_page -= self.center().y()
        y_on_page *= self.height_scale_factor()
        return y_on_page + self.geometry.height() / 2

    def unproject_point(self, point) -> Point2D:
        return Point2D(self.unproject_x(point.x()), self.unproject_y(point.y()))

    def unproject_rect(self, rect: Rectangle) -> Rectangle:
        top_left = self.unproject_point(rect.top_left())
        bottom_right = self.unproject_point(rect.bottom_right())
        return Rectangle.from_points(top_left, bottom_right)

    def unproject_x(self, x_on_screen: float) -> float:
        x_on_screen -= self.geometry.width() / 2
        x_on_screen /= self.height_scale_factor()
        return x_on_screen + self.center().x()

    def unproject_y(self, y_on_screen: float) -> float:
        y_on_screen -= self.geometry.height() / 2
        y_on_screen /= self.height_scale_factor()
        return y_on_screen + self.center().y()



import traceback
from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication, QMessageBox

import fusion
from fusion.logging import LoggingLevels, LOGGING_LEVEL
import pamet
from pamet.views.selector_widget import SelectorWidget

log = fusion.get_logger(__name__)


class DesktopApp(QApplication):
    def __init__(self):
        QApplication.__init__(self)

        name = 'pamet'

        if LOGGING_LEVEL == LoggingLevels.DEBUG:
            name += '-debug'

        self.setOrganizationName('p10')
        self.setApplicationName(name)
        self.setApplicationVersion(pamet.__version__)
        self.setQuitOnLastWindowClosed(True)  # Needed for some reason

        # TODO: This is a hack to make the app not crash when showing
        # a message box from a command
        self.message_box = QMessageBox()

        # Selector widget for grabbing screen snipptes
        self.selector_widget = SelectorWidget()

    def hacky_present_modal(self, dialog):
        # Otherwise the app crashes when showing the dialog from a command
        def present():
            dialog.exec()

        QTimer.singleShot(0, present)

    def present_exception(self, exception: Exception, title: str = None):
        title = title or 'Exception raised'
        text = str(exception) + '\n\nTraceback:\n' + traceback.format_exc()
        self.message_box.setText(text)
        self.message_box.setWindowTitle(title)
        self.message_box.setIcon(QMessageBox.Critical)
        self.hacky_present_modal(self.message_box)
        # QMessageBox.critical(self.activeWindow(), title, text)



from __future__ import annotations
from PySide6.QtGui import QCloseEvent
from PySide6.QtNetwork import QNetworkAccessManager, QNetworkReply

from PySide6.QtWidgets import QCompleter, QPushButton

import fusion
from pamet import register_note_view_type

import pamet
from pamet.desktop_app import icons
from pamet.util.url import Url
from pamet.model.card_note import CardNote
from pamet.model.image_note import ImageNote
from pamet.model.page import Page
from pamet.model.text_note import TextNote
from pamet.views.note.image.props_edit.widget import ImagePropsWidget
from pamet.views.note.anchor.edit_widget import AnchorEditWidgetMixin
from pamet.views.note.anchor.props_edit.widget import AnchorEditPropsWidget
from pamet.views.note.base_edit.view_state import NoteEditViewState
from pamet.views.note.base_edit.widget import BaseNoteEditWidget
from pamet.views.note.card.props_edit.widget import TextEditPropsWidget

log = fusion.get_logger(__name__)


class CardEditViewState(NoteEditViewState):
    pass


@register_note_view_type(state_type=CardEditViewState,
                         note_type=(TextNote, ImageNote, CardNote),
                         edit=True)
class CardNoteEditWidget(BaseNoteEditWidget, AnchorEditWidgetMixin):

    def __init__(self, parent, initial_state: CardEditViewState):
        super().__init__(parent, initial_state)
        self._network_am = QNetworkAccessManager(self)
        self._image_download_reply: QNetworkReply = None

        # Add the link props widget
        self.text_props_widget = TextEditPropsWidget(self)
        self.image_props_widget = ImagePropsWidget(self)
        self.link_props_widget = AnchorEditPropsWidget(self)

        self.ui.centralAreaWidget.layout().addWidget(self.link_props_widget)
        self.ui.centralAreaWidget.layout().addWidget(self.image_props_widget)
        self.ui.centralAreaWidget.layout().addWidget(self.text_props_widget)

        # Setup the link props widget
        self.link_props_widget.ui.pametPageLabel.hide()
        self.link_props_widget.ui.invalidUrlLabel.hide()
        self.link_props_widget.ui.urlLineEdit.textChanged.connect(
            self.on_url_change)

        # Setup a completer with all the page names except the current one
        self_page = pamet.page(initial_state.page_id)
        page_completer = QCompleter(
            (p.name for p in pamet.pages() if p != self_page), parent=self)
        self.link_props_widget.ui.urlLineEdit.setCompleter(page_completer)

        # Setup the text props widget

        if isinstance(self.edited_note, (TextNote, CardNote)):
            self.text_edit.setPlainText(self.edited_note.text)

        # Add toolbar actions
        self.text_button = QPushButton(icons.text, '', self)
        self.text_button.setCheckable(True)
        self.ui.toolbarLayout.insertWidget(0, self.text_button)

        self.image_button = QPushButton(icons.image, '', self)
        self.image_button.setCheckable(True)
        self.ui.toolbarLayout.insertWidget(1, self.image_button)

        self.link_button = QPushButton(icons.link, '', self)
        self.link_button.setCheckable(True)
        self.ui.toolbarLayout.insertWidget(2, self.link_button)

        # Setup the buttons according to the type of the edited note
        if isinstance(self.edited_note, ImageNote):
            self.image_button.setChecked(True)
        else:
            self.image_props_widget.hide()

        if isinstance(self.edited_note, TextNote):
            self.text_button.setChecked(True)
        else:
            self.text_props_widget.hide()
        # The card note inherits both, so if it's a card - both will be checked

        # Setup the link button and line edit
        if initial_state.url.is_empty():
            self.link_props_widget.hide()
            self.text_props_widget.ui.getTitleButton.hide()
        else:
            self.link_button.setChecked(True)
            self.link_props_widget.ui.urlLineEdit.setText(
                str(initial_state.url))

        self.text_button.toggled.connect(self.handle_text_button_toggled)
        self.image_button.toggled.connect(self.handle_image_button_toggled)
        self.link_button.toggled.connect(self.handle_link_button_toggled)
        self.update_buttons_from_note_type()

    @property
    def text_edit(self):
        return self.text_props_widget.text_edit

    # This shouldn't be specified explicitly IMO, but I couldn't find the bug
    def focusInEvent(self, event) -> None:
        edited_note = self.edited_note

        if not self.edited_note.url.is_empty():
            self.link_props_widget.ui.urlLineEdit.setFocus()
        elif isinstance(edited_note, TextNote):
            self.text_edit.setFocus()
        elif isinstance(edited_note, ImageNote):
            self.image_props_widget.ui.urlLineEdit.setFocus()
        elif isinstance(edited_note, CardNote):
            self.text_edit.setFocus()

    def update_note_type_from_buttons(self):
        if self.text_button.isChecked() and self.image_button.isChecked():
            self.edited_note = CardNote(**self.edited_note.asdict())
        elif self.text_button.isChecked():
            self.edited_note = TextNote(**self.edited_note.asdict())
        elif self.image_button.isChecked():
            self.edited_note = ImageNote(**self.edited_note.asdict())
        else:
            raise Exception

        self.update_buttons_from_note_type()

    def update_buttons_from_note_type(self):
        # Should show/hide the widget and if toggled off and the text is not
        # shown - show it (i.e. revert to basic text note)
        if not (self.text_button.isChecked() or self.image_button.isChecked()):
            self.text_button.setChecked(True)

        # Don't allow removing the text if there's no image
        elif self.text_button.isChecked() and \
                not self.image_button.isChecked():
            self.text_button.setEnabled(False)
        else:
            self.text_button.setEnabled(True)

    def handle_text_button_toggled(self, checked: bool):
        if checked:
            self.text_props_widget.show()
        else:
            self.text_props_widget.hide()
        self.update_buttons_from_note_type()
        self.update_note_type_from_buttons()

    def handle_image_button_toggled(self, checked: bool):
        if checked:
            self.image_props_widget.show()
            self.image_props_widget.ui.urlLineEdit.setFocus()

            if not self.text_edit.toPlainText():
                self.text_button.setChecked(False)  # This should call the hide
        else:
            self.image_props_widget.hide()
        self.update_buttons_from_note_type()
        self.update_note_type_from_buttons()

    def handle_link_button_toggled(self, checked: bool):
        if checked:
            self.link_props_widget.show()
            self.link_props_widget.ui.urlLineEdit.setFocus()
            self.text_props_widget.ui.getTitleButton.show()
        else:
            self.link_props_widget.hide()
            self.edited_note.url = None
            self.text_props_widget.ui.getTitleButton.hide()

    def decorate_for_internal_url(self, url_is_internal):
        if url_is_internal:
            self.link_props_widget.ui.pametPageLabel.show()
            self.text_edit.setEnabled(False)
        else:
            self.link_props_widget.ui.pametPageLabel.hide()
            self.text_edit.setEnabled(True)

    def on_url_change(self):
        url_field = self.link_props_widget.ui.urlLineEdit.text()
        self.edited_note.url = url_field

        # If there's no scheme it's probably an internal page (or a web link)
        url = Url(url_field)
        if not url.scheme:
            # We check if it happens to be the name of a note page
            page_by_name = None
            for page in pamet.pages():
                if page.name == url_field:
                    page_by_name = page
                    break

            # If there's a page with that name setup the url accordingly
            if page_by_name:
                self.decorate_for_internal_url(True)
                self.edited_note.url = page_by_name.url()
                self.text_edit.setText(page_by_name.name)
                self.edited_note.text = page_by_name.name
            # Else assume it's a web link without a schema
            else:
                self.decorate_for_internal_url(False)

        else:  # If there is a schema
            # If an URI is pasted with a pamet schema
            linked_page: Page = pamet.page(self.edited_note.url.get_page_id())
            if linked_page:
                self.decorate_for_internal_url(True)
                self.link_props_widget.ui.urlLineEdit.setText(linked_page.name)
                return

            # If it's clearly a web URL and the user has set the text -
            # get the title
            if url.scheme in ['http', 'https'] and \
                    not self.text_edit.toPlainText():
                self.text_props_widget._get_title()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self._image_download_reply:
            self._image_download_reply.abort()
        return super().closeEvent(event)



from typing import Union
from fusion.util.point2d import Point2D
from fusion.util.rectangle import Rectangle
from fusion.libs.entity import entity_type
from pamet.util.url import Url
from pamet.model.note import Note

IMAGE_URL = 'image_url'
LOCAL_IMAGE_URL = 'local_image_url'
IMAGE_SIZE = 'image_size'
IMAGE_MD5 = 'image_md5'


@entity_type
class ImageNote(Note):

    @property
    def image_url(self) -> Url:
        return Url(self.content.get(IMAGE_URL, ''))

    @image_url.setter
    def image_url(self, new_url: Union[Url, str, None]):
        if isinstance(new_url, Url):
            new_url = str(new_url)
        if not new_url:
            self.content.pop(IMAGE_URL, None)
            return
        self.content[IMAGE_URL] = new_url

    @property
    def local_image_url(self) -> Url:
        return Url(self.content.get(LOCAL_IMAGE_URL, ''))

    @local_image_url.setter
    def local_image_url(self, new_url: Union[Url, str, None]):
        if isinstance(new_url, Url):
            new_url = str(new_url)
        if not new_url:
            self.content.pop(LOCAL_IMAGE_URL, None)
            return
        self.content[LOCAL_IMAGE_URL] = new_url

    @property
    def image_size(self) -> Point2D:
        if IMAGE_SIZE in self.metadata:
            return Point2D(*self.metadata[IMAGE_SIZE])
        return None

    @image_size.setter
    def image_size(self, size: Point2D):
        self.metadata[IMAGE_SIZE] = size.as_tuple()

    @property
    def image_md5(self) -> Point2D:
        if IMAGE_MD5 in self.metadata:
            return self.metadata[IMAGE_MD5]
        return None

    @image_md5.setter
    def image_md5(self, md5sum: Point2D):
        self.metadata[IMAGE_MD5] = md5sum

    def image_rect(self) -> Rectangle:
        note_rect: Rectangle = self.rect()
        image_AR = self.image_size.x() / self.image_size.y()
        note_AR = note_rect.width() / note_rect.height()

        if note_AR > image_AR:
            width = image_AR * note_rect.height()
            height = note_rect.height()
            return Point2D(width, height)
        else:
            width = note_rect.width()
            height = note_rect.width() / image_AR
            return Point2D(width, height)



from PySide6.QtWidgets import QMessageBox, QWidget
from fusion.libs.entity.change import Change

from fusion.libs.state import ViewState, view_state_type
from fusion.platform.qt_widgets import bind_and_apply_state
from fusion.view import View

import pamet
from pamet import actions
from pamet.model.page import Page
from .ui_properties_widget import Ui_MapPagePropertiesWidget


@view_state_type
class MapPagePropertiesViewState(ViewState):
    focused_prop: str = ''

    @property
    def page_id(self):
        return self.id


class MapPagePropertiesWidget(QWidget, View):

    def __init__(self, tab_widget, initial_state):
        QWidget.__init__(self, parent=tab_widget)
        View.__init__(self, initial_state=initial_state)

        self.ui = Ui_MapPagePropertiesWidget()
        self.ui.setupUi(self)

        self.parent_tab = tab_widget

        # Configure the UI
        # Set the home page check box state
        default_page = pamet.sync_repo().default_page()
        if default_page and default_page.id == initial_state.page_id:
            self.ui.setAsHomePageCheckBox.setChecked(True)

        self.ui.nameWarningLabel.hide()
        self.ui.nameWarningLabel.setStyleSheet(
            'QLabel {color : red; font-size: 8pt;}')

        # Bind UI signals to the appropriate methods
        self.ui.nameLineEdit.textChanged.connect(
            self._handle_name_line_edit_text_changed)

        self.ui.saveButton.clicked.connect(self._save_page)
        self.ui.setAsHomePageCheckBox.toggled.connect(
            self._handle_home_page_toggle)
        self.ui.deleteButton.clicked.connect(self._handle_delete_button_click)
        self.ui.deleteButton.setStyleSheet('QButton {background-color: red;}')

        bind_and_apply_state(self, initial_state, self.on_state_change)

    def on_state_change(self, change: Change):
        # Fill the prop fields with the respective values
        if change.updated.page_id:
            page = pamet.page(self.state().page_id)
            self.ui.nameLineEdit.setText(page.name)

        # Apply the field focus when specified
        if change.updated.focused_prop:
            if self.state().focused_prop == 'name':
                self.ui.nameLineEdit.selectAll()

    def _handle_delete_button_click(self):
        reply = QMessageBox.question(
            self, 'Delete page', 'Are you sure you want to delete the page?')
        if reply == QMessageBox.StandardButton.Yes:
            actions.map_page.delete_page(self.parent_tab.state(),
                                         pamet.page(self.state().page_id))
        actions.tab.close_page_properties(self.parent_tab.state())

    def _save_page(self):
        state = self.state()
        page = pamet.page(state.page_id)
        page.name = self.ui.nameLineEdit.text()
        actions.map_page.save_page_properties(page)
        actions.tab.close_page_properties(self.parent_tab.state())

    def _handle_name_line_edit_text_changed(self, new_text):
        # Check if the id is available and valid
        state = self.state()
        if new_text == pamet.page(state.page_id).name:
            self.ui.nameWarningLabel.hide()
            self.ui.saveButton.setEnabled(True)
            return

        if not new_text:
            self.ui.nameWarningLabel.setText('The page name can not be empty')
            self.ui.nameWarningLabel.show()
            self.ui.saveButton.setEnabled(False)
            return

        elif pamet.find_one(name=new_text, type=Page):
            self.ui.nameWarningLabel.setText('Name already taken')
            self.ui.nameWarningLabel.show()
            self.ui.saveButton.setEnabled(False)
            return

        self.ui.saveButton.setEnabled(True)
        self.ui.nameWarningLabel.hide()

    def _handle_home_page_toggle(self, checked: bool):
        if checked:
            home_page = pamet.page(self.state().page_id)
        else:
            home_page = None
        pamet.sync_repo().set_default_page(home_page)



import os
from pathlib import Path
from PySide6.QtWidgets import QMessageBox
import click

import fusion
from fusion.libs.action import actions_log_channel
from fusion.libs.action.action_call import ActionCall, ActionRunStates
from fusion.logging import LOGGING_LEVEL, LoggingLevels
from pamet import channels as pamet_channels, commands, set_semantic_search_service

import pamet
from pamet import desktop_app
from pamet.actions import window as window_actions
from pamet.actions import other as other_actions
from pamet.desktop_app.app import DesktopApp
from pamet.desktop_app.init_config import configure_for_qt
from pamet.model.page import Page

from pamet.services.backup import AnotherServiceAlreadyRunningException
from pamet.services.backup import FSStorageBackupService
from pamet.services.local_server import LocalServer
from pamet.services.other_pages_list_update import OtherPagesListUpdateService

from pamet.services.search.fuzzy import FuzzySearchService
from pamet.services.search.semantic import SemanticSearchService
from pamet.services.undo import UndoService
from pamet.storage import FSStorageRepository
from pamet.views.window.widget import WindowWidget

log = fusion.get_logger(__name__)


def raise_a_window():
    windows = [
        w for w in DesktopApp.instance().topLevelWidgets()
        if isinstance(w, WindowWidget)
    ]
    if windows:
        windows[0].show()
        windows[0].activateWindow()
        windows[0].raise_()


local_server_commands = {
    'grab_screen_snippet': commands.grab_screen_snippet,
    'raise_window': raise_a_window
}


@click.command()
@click.argument('path', type=click.Path(exists=True), required=False)
@click.option('--command', type=click.Choice(local_server_commands.keys()))
def main(path: str, command: str):
    # Check if another instance is running
    local_server = LocalServer(commands=local_server_commands)
    if local_server.another_instance_is_running():
        port = local_server.get_port_from_lock_file()
        if command:
            LocalServer.send_command(port, command)
        else:
            LocalServer.send_command(port, 'raise_window')
        return
    else:
        local_server.start()

    app = DesktopApp()
    app.aboutToQuit.connect(local_server.stop)

    configure_for_qt(app)
    # Load the config
    user_config = desktop_app.get_user_settings()
    # If there's changes after the load it means that some default is not saved
    # to disk or some other irregularity has been handled by the config class
    if user_config.changes_present:
        desktop_app.save_user_settings(user_config)

    # Init the repo
    if path:
        repo_path = Path(path)
    else:
        repo_path = Path(user_config.repository_path)
    log.info('Using repository: %s' % repo_path)

    # Testing - restore legacy backups

    # # V2
    # v2_backup_folder = repo_path / '__v2_legacy_pages_backup__'
    # for file in v2_backup_folder.iterdir():

    #     v3_file_in_repo = file.parent.parent / file.name
    #     v3_file_in_repo = v3_file_in_repo.with_suffix('').with_suffix('.json')

    #     if v3_file_in_repo.exists():
    #         v3_file_in_repo.unlink()
    #     restore_path = file.parent.parent / file.name
    #     restore_path = restore_path.with_suffix('')
    #     file.rename(restore_path)

    # for file in repo_path.iterdir():
    #     if file.name.endswith('.pam4.json'):
    #         file.unlink()

    # # V3
    # v3_backup_folder = repo_path / '__v3_legacy_pages_backup__'
    # backup_page_names = []
    # new_paths_by_old_path = {}
    # for file in v3_backup_folder.iterdir():
    #     if file.name.endswith('.backup'):
    #         new_path = file.parent.parent / file.with_suffix('').name
    #         new_paths_by_old_path[file] = new_path

    #         page_data = json.loads(file.read_text())
    #         page_name = file.with_suffix('').with_suffix('').name
    #         backup_page_names.append(page_name)

    # # Read all page names in the repo
    # page_paths_by_name = {}
    # for file in repo_path.iterdir():
    #     if file.is_file() and file.name.endswith('.pam4.json'):
    #         page_data = json.loads(file.read_text())
    #         name = page_data['name']
    #         page_paths_by_name[name] = file

    # # Delete all previously imported (by name matching)
    # for page_name in backup_page_names:
    #     if page_name in page_paths_by_name:
    #         page_paths_by_name[page_name].unlink()
    #     else:
    #         pass

    # # Restore the backups
    # for old_path, new_path in new_paths_by_old_path.items():
    #     old_path.rename(new_path)

    if os.path.exists(repo_path):
        fs_repo = FSStorageRepository.open(repo_path,
                                           queue_save_on_change=True)
        legacy_page_paths = fs_repo.process_legacy_pages()
        fs_repo.load_all_pages()

        # Checksum legacy pages and fix internal links in them
        for page_path in legacy_page_paths:
            page_id = fs_repo.id_from_page_path(page_path)
            page = fs_repo.find_one(id=page_id)

            fs_repo.checksum_imported_page_notes(page)
            fs_repo.fix_legacy_page_internal_links(page)
    else:
        fs_repo = FSStorageRepository.new(repo_path, queue_save_on_change=True)

    repo_settings = desktop_app.get_repo_settings(repo_path)
    if repo_settings.changes_present:
        desktop_app.save_repo_settings(repo_settings)

    pamet.set_sync_repo(fs_repo)
    pamet.set_undo_service(
        UndoService(pamet_channels.entity_change_sets_per_TLA))

    # There should be a better place to do call the validity checks I guess
    # pamet.model.arrow_validity_check()

    # # Debug
    # misli_channels.state_changes_per_TLA_by_id.subscribe(
    #     lambda x: print(f'STATE_CHANGES_BY_ID CHANNEL: {x}'))

    start_page = pamet.default_page() or fs_repo.find_one(type=Page)
    if not start_page:
        start_page = other_actions.create_default_page()

    window_state = window_actions.new_browser_window()
    window = WindowWidget(initial_state=window_state)
    window.showMaximized()

    # fusion.call_delayed(window_actions.new_browser_tab,
    #                    args=[window_state, start_page])
    window_actions.new_browser_tab(window_state, start_page)

    search_service = FuzzySearchService(
        pamet_channels.entity_change_sets_per_TLA)
    search_service.load_all_content()
    pamet.set_search_service(search_service)

    other_page_list_service = OtherPagesListUpdateService()
    other_page_list_service.start()

    # Setup exception reporting for failed actions
    if LOGGING_LEVEL != LoggingLevels.DEBUG.value:

        def show_exception_for_failed_action(action_call: ActionCall):
            if action_call.run_state != ActionRunStates.FAILED:
                return
            title = f'Exception raised during action "{action_call.name}"'
            app.present_exception(exception=action_call.error, title=title)

        actions_log_channel.subscribe(show_exception_for_failed_action)

    if repo_settings.backups_enabled:
        backup_service = FSStorageBackupService(
            backup_folder=repo_settings.backup_folder,
            repository=fs_repo,
            changeset_channel=pamet_channels.entity_change_sets_per_TLA,
            record_all_changes=repo_settings.record_all_changes)

        service_started = False
        try:
            backup_service.start()
            service_started = True
        except AnotherServiceAlreadyRunningException:
            log.info('Backup service not started. '
                     'Probably another instance is running')
            reply = QMessageBox.question(
                window, 'Backup service conflict',
                'A backup service lock is present. If you\'re sure there\'s '
                'no other instances running on the same repo - '
                'press Yes to override.')
            if reply == QMessageBox.StandardButton.Yes:
                backup_service.service_lock_path().unlink()
                backup_service.start()
                service_started = True

        if service_started:
            app.aboutToQuit.connect(backup_service.stop)

    # Experimental semantic search
    if repo_settings.semantic_search_enabled:
        semantic_search_service = SemanticSearchService(
            data_folder=repo_path / '__semantic_index__',
            change_set_channel=pamet_channels.entity_change_sets_per_TLA)

        print('Loading semantic search index...')
        semantic_search_service.load_all_content()
        print('Semantic search index loaded')
        set_semantic_search_service(semantic_search_service)

    fusion.set_main_loop_exception_handler(
        lambda e: app.present_exception(e, title='Main loop exception')
    )

    return app.exec()


if __name__ == '__main__':
    main()



from __future__ import annotations

from pamet.model.op_list_note import OtherPageListNote
from pamet.note_view_lib import register_note_view_type
from pamet.views.note.base_edit.view_state import NoteEditViewState
from pamet.views.note.base_edit.widget import BaseNoteEditWidget


@register_note_view_type(state_type=NoteEditViewState,
                         note_type=OtherPageListNote,
                         edit=True)
class OtherPageListNoteEditWidget(BaseNoteEditWidget):
    pass
    # def __init__(self, parent: TabWidget, initial_state: NoteEditViewState):
    #     super().__init__(parent, initial_state)



from PySide6.QtCore import Qt
from PySide6.QtGui import QImage, QPixmap, QResizeEvent
from PySide6.QtWidgets import QLabel
from pathlib import Path

from pamet.desktop_app import media_store
from pamet.util.url import Url


class ImageLabel(QLabel):

    def __init__(self, parent):
        QLabel.__init__(self, parent)
        self._image = None
        self._old_size = self.size()

        self.setWordWrap(True)
        self.setAlignment(Qt.AlignCenter)

    def image(self) -> QImage:
        return self._image

    def update_image_cache(self, url: Url, local_url: Url):
        if local_url:
            local_path = Path(local_url.path)
            if local_url.is_internal():
                local_path = media_store().path_for_internal_uri(local_url)

            if not local_path.exists():
                self.setText(f'Local path missing: {local_path}')
            else:
                self._image = QImage(local_path)
                if self._image.isNull():
                    self.setText(f'Could not load the image at {local_path}')
                    self._image = None
                    return
                self.update_pixmap()

        else:
            if not url:
                self.setText('No image chosen')
            else:
                self.setText(
                    'Image was not properly downloaded. Edit it to try again.')

    def update_pixmap(self):
        if not self._image:
            return
        scaled_image = self._image.scaled(self.size(), Qt.KeepAspectRatio,
                                          Qt.SmoothTransformation)
        self.setPixmap(QPixmap.fromImage(scaled_image))

    def resizeEvent(self, event: QResizeEvent):
        new_size = self.rect().size()
        if new_size != self._old_size:
            self._old_size = new_size
            self.update_pixmap()




import random
from fusion import get_logger
from fusion import entity_type
from fusion.util.point2d import Point2D
from fusion.util.rectangle import Rectangle
from pamet.constants import NOTE_MARGIN
from pamet.model.note import Note
log = get_logger(__name__)

RANDOMIZE_TEXT = False

TEXT = 'text'
letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
letters += 'абвгдежзийклмнопрстуфхцчшщьъюяАБВГДЕЖЗИЙКЛМНОПРСТУФХЦЧШЩѝЪЮЯ'


@entity_type
class TextNote(Note):
    def __post_init__(self):
        super().__post_init__()
        if TEXT not in self.content:
            self.content[TEXT] = ''

    @property
    def text(self) -> str:
        try:
            text = self.content[TEXT]
        except KeyError:
            text = ''
            self.content[TEXT] = text

        if RANDOMIZE_TEXT:
            new_text = ''
            for char in text:
                if char != ' ':
                    char = random.choice(letters)
                new_text += char
            text = new_text
        return text

    @text.setter
    def text(self, new_text: str):
        self.content[TEXT] = new_text

    def text_rect(self, for_size: Point2D = None) -> Rectangle:
        if for_size:
            size = for_size
        else:
            size = self.rect().size()
        size -= Point2D(2 * NOTE_MARGIN, 2 * NOTE_MARGIN)
        return Rectangle(NOTE_MARGIN, NOTE_MARGIN, *size.as_tuple())



from pathlib import Path
from fusion.visual_inspection.action_camera import ActionCamera


def test_record_new(window_fixture):
    rec_path = Path(__file__).parent / 'test_new.py'
    ActionCamera().record_until_exit(rec_path)
    assert rec_path.exists()



from __future__ import annotations

from typing import Generator, List, Union
from dataclasses import field
from datetime import datetime

from fusion import get_logger
from fusion import entity_type
from fusion.util import Point2D, Rectangle, Color

import pamet
from pamet.constants import DEFAULT_NOTE_HEIGHT, DEFAULT_NOTE_WIDTH
from pamet.constants import DEFAULT_BG_COLOR, DEFAULT_COLOR
from pamet.constants import MIN_NOTE_WIDTH, MIN_NOTE_HEIGHT
from pamet.constants import MAX_NOTE_WIDTH, MAX_NOTE_HEIGHT
from pamet.model.page_child import PageChild
from pamet.util.url import Url
from fusion.util import current_time, timestamp
from pamet.model.arrow import Arrow, ArrowAnchorType

log = get_logger(__name__)

URL = 'url'


@entity_type
class Note(PageChild):
    geometry: list = field(default_factory=lambda:
                           [0, 0, DEFAULT_NOTE_WIDTH, DEFAULT_NOTE_HEIGHT])
    style: dict = field(default_factory=dict)
    content: dict = field(default_factory=dict)
    metadata: dict = field(default_factory=dict)
    created: str = field(default_factory=lambda: timestamp((current_time())))
    modified: str = field(default_factory=lambda: timestamp((current_time())))
    tags: List[str] = field(default_factory=list)

    def __repr__(self):
        return f'<{type(self).__name__} gid={self.gid()}>'

    def asdict(self):
        self_dict = super().asdict()
        return self_dict

    def parent_gid(self):
        return self.page_id

    def rect(self) -> Rectangle:
        return Rectangle(*self.geometry)

    @property
    def color(self):
        return self.style.get('color', DEFAULT_COLOR)

    @color.setter
    def color(self, new_color: tuple | Color):
        if isinstance(new_color, Color):
            new_color = new_color.as_tuple()

        if new_color is None:
            self.style.pop('color', None)
            return
        self.style['color'] = new_color

    @property
    def background_color(self):
        return self.style.get('background_color', DEFAULT_BG_COLOR)

    @background_color.setter
    def background_color(self, new_background_color: tuple | background_color):
        if isinstance(new_background_color, Color):
            new_background_color = new_background_color.as_tuple()

        if new_background_color is None:
            self.style.pop('background_color', None)
            return
        self.style['background_color'] = new_background_color

    def set_color(self, color: Color):
        self.color = color.as_tuple()

    def get_color(self) -> Color:
        return Color(*self.color)

    def get_background_color(self) -> Color:
        return Color(*self.background_color)

    @property
    def width(self) -> float:
        return self.geometry[2]

    @width.setter
    def width(self, width: float) -> float:
        width = min(MAX_NOTE_WIDTH, max(width, MIN_NOTE_WIDTH))
        self.geometry[2] = width

    @property
    def height(self) -> float:
        return self.geometry[3]

    @height.setter
    def height(self, height: float):
        height = min(MAX_NOTE_HEIGHT, max(height, MIN_NOTE_HEIGHT))
        self.geometry[3] = height

    @property
    def x(self) -> float:
        return self.geometry[0]

    @x.setter
    def x(self, x: float):
        self.geometry[0] = x

    @property
    def y(self) -> float:
        return self.geometry[1]

    @y.setter
    def y(self, y: float) -> float:
        self.geometry[1] = y

    def size(self) -> Point2D:
        return Point2D(self.width, self.height)

    def set_size(self, new_size: Point2D):
        self.width = new_size.x()
        self.height = new_size.y()

    def set_rect(self, new_rect: Rectangle):
        self.x = new_rect.x()
        self.y = new_rect.y()
        self.width = new_rect.width()
        self.height = new_rect.height()

    @property
    def url(self) -> Url:
        return Url(self.content.get(URL, ''))

    @url.setter
    def url(self, new_url: Union[Url, str, None]):
        if isinstance(new_url, Url):
            new_url = str(new_url)
        # If None is passed - remove the URL parameter (may be unnecessary)
        elif new_url is None:
            self.content.pop(URL, None)
            return
        self.content[URL] = new_url

    def in_arrows(self) -> Generator[Arrow, None, None]:
        for arrow in pamet.arrows(self.page_id):
            if self.own_id == arrow.head_note_id:
                yield arrow

    def out_arrows(self) -> Generator[Arrow, None, None]:
        for arrow in pamet.arrows(self.page_id):
            if self.own_id == arrow.tail_note_id:
                yield arrow

    def connected_arrows(self) -> Generator[Arrow, None, None]:
        for arrow in pamet.arrows(self.page_id):
            if self.own_id == arrow.tail_note_id or \
                    self.own_id == arrow.head_note_id:
                yield arrow

    def arrow_anchor(self, anchor_type: ArrowAnchorType) -> Point2D:
        """Returns the center position for a specific arrow anchor.

        Raises: Does not support the ArrowAnchorType.AUTO
        """
        rect: Rectangle = self.rect()

        match anchor_type:
            case ArrowAnchorType.MID_LEFT:
                return rect.top_left() + Point2D(0, rect.height() / 2)
            case ArrowAnchorType.TOP_MID:
                return rect.top_left() + Point2D(rect.width() / 2, 0)
            case ArrowAnchorType.MID_RIGHT:
                return rect.top_right() + Point2D(0, rect.height() / 2)
            case ArrowAnchorType.BOTTOM_MID:
                return rect.bottom_left() + Point2D(rect.width() / 2)
            case _:
                raise Exception

    @property
    def datetime_created(self) -> datetime:
        return datetime.fromisoformat(self.created)

    @datetime_created.setter
    def datetime_created(self, new_dt: datetime):
        self.created = timestamp(new_dt)

    @property
    def datetime_modified(self) -> datetime:
        return datetime.fromisoformat(self.modified)

    @datetime_modified.setter
    def datetime_modified(self, new_dt: datetime):
        self.modified = timestamp(new_dt)



from __future__ import annotations

from dataclasses import field
from enum import Enum
from typing import List

import pamet
from fusion.util import Rectangle, Point2D
from fusion.libs.state import view_state_type, ViewState
from pamet.constants import DEFAULT_EYE_HEIGHT
from pamet.util.url import Url
from pamet.model.note import Note
from pamet.model.page import Page
from pamet.views.arrow.widget import ArrowViewState
from pamet.views.map_page.viewport import Viewport
from pamet.views.note.base.state import NoteViewState


class MapPageMode(Enum):
    NONE = 0
    DRAG_NAVIGATION = 1
    DRAG_SELECT = 2
    NOTE_RESIZE = 3
    CHILD_MOVE = 4
    CREATE_ARROW = 5
    ARROW_EDGE_DRAG = 6


@view_state_type
class MapPageViewState(ViewState, Viewport, Page):
    # page_id: str = ''
    note_view_states: set[NoteViewState] = field(default_factory=set)
    arrow_view_states: set[ArrowViewState] = field(default_factory=set)

    geometry: Rectangle = Rectangle(0, 0, 500, 500)
    viewport_center: Point2D = field(default_factory=Point2D)
    viewport_height: float = DEFAULT_EYE_HEIGHT

    special_mode: int = 0

    # drag_navigation_active: bool = False
    drag_navigation_start_position: Point2D = None

    selected_children: set = field(default_factory=set)

    # Field/drag selection
    mouse_position_on_drag_select_start: Point2D = None
    drag_selected_children: set = field(default_factory=set)
    drag_select_rect_props: list = field(default_factory=list)

    # Note resize related
    mouse_position_on_note_drag_start: Point2D = None
    note_resize_click_position: Point2D = None
    note_resize_delta_from_note_edge: Point2D = None
    note_resize_main_note: Note = None
    viewport_position_on_press: Point2D = None
    note_resize_states: List[NoteViewState] = field(default_factory=list)

    # Note move related
    moved_note_states: List[NoteViewState] = field(default_factory=list)
    moved_arrow_states: List[ArrowViewState] = field(default_factory=list)

    # Arrow manipulation related
    new_arrow_view_states: List[ArrowViewState] = field(default_factory=list)
    arrow_with_visible_cps: ArrowViewState = None
    dragged_edge_index: float = None

    def __repr__(self):
        return (f'<MapPageViewState page_id={self.page_id}'
                f' {len(self.note_view_states)=}>')

    @property
    def page_id(self):
        return self.id

    def mode(self):
        return MapPageMode(self.special_mode)

    def set_mode(self, mode: MapPageMode):
        if self.mode() != MapPageMode.NONE:
            self.clear_mode()
        self.special_mode = mode.value

    def clear_mode(self):
        self.special_mode = MapPageMode.NONE
        self.drag_navigation_start_position = None
        self.mouse_position_on_drag_select_start = None
        self.drag_selected_children = set()
        self.drag_select_rect_props = []

        self.mouse_position_on_note_drag_start = None
        self.note_resize_click_position = None
        self.note_resize_delta_from_note_edge = None
        self.note_resize_main_note = None
        self.viewport_position_on_press = None
        self.note_resize_states.clear()

        self.moved_note_states.clear()
        self.moved_arrow_states.clear()

        self.new_arrow_view_states.clear()

    def view_state_for_note_own_id(self, note_own_id: str):
        for note_vs in self.note_view_states:
            if note_vs.note_gid == (self.page_id, note_own_id):
                return note_vs
        return None

    def url(self):
        return Url(f'pamet:///p/{self.page_id}')

    def page_url(self, with_anchor=True) -> Url:
        url = self.url()
        center = self.viewport_center

        if with_anchor:
            url = url.with_anchor(eye_pos=center,
                                  eye_height=self.viewport_height)

        return url



from time import sleep
from pamet.services.local_server import LocalServer


def test_local_server(tmp_path):

    # Start the server
    server1 = LocalServer(config_dir=tmp_path)
    server2 = LocalServer(config_dir=tmp_path)
    server1.start()
    sleep(0.1)
    # Check that it's running
    assert server2.another_instance_is_running()

    # Stop the server
    server1.stop()

    # Check that it's not running
    assert not server2.another_instance_is_running()
